#!/usr/bin/env python

"""
WorkQueueRegMgrInterface test
"""

import unittest

from WMCore_t.WorkQueue_t.WorkQueueTestCase import WorkQueueTestCase
from nose.plugins.attrib import attr

from WMCore.WorkQueue.WorkQueue import globalQueue, localQueue
from WMCore.WorkQueue.WorkQueueReqMgrInterface import WorkQueueReqMgrInterface
from WMQuality.Emulators.DataBlockGenerator.Globals import GlobalParams
from WMQuality.Emulators.ReqMgrClient.ReqMgr import ReqMgr as fakeReqMgr


def getFirstTask(wmspec):
    """Return the 1st top level task"""
    # http://www.logilab.org/ticket/8774
    # pylint: disable=E1101,E1103
    return next(wmspec.taskIterator())


class WorkQueueReqMgrInterfaceTest(WorkQueueTestCase):
    """
    TestCase for WorkQueueReqMgrInterface module
    """

    def setSchema(self):
        self.schema = []
        self.couchApps = ["WorkQueue"]

    def setUp(self):
        self.reqmgr2_endpoint = "https://cmsweb-testbed.cern.ch/reqmgr2"

        WorkQueueTestCase.setUp(self)
        GlobalParams.resetParams()
        self.globalQCouchUrl = "%s/%s" % (self.testInit.couchUrl, self.globalQDB)
        self.localQCouchUrl = "%s/%s" % (self.testInit.couchUrl,
                                         self.localQInboxDB)

    def tearDown(self):
        WorkQueueTestCase.tearDown(self)
        GlobalParams.resetParams()

    def getConfig(self):
        """
        _createConfig_

        General config file
        """
        # configPath=os.path.join(WMCore.WMInit.getWMBASE(), \
        #                        'src/python/WMComponent/WorkQueueManager/DefaultConfig.py')):

        config = self.testInit.getConfiguration()
        # http://www.logilab.org/ticket/8961
        # pylint: disable=E1101, E1103
        config.component_("WorkQueueManager")
        config.section_("General")
        config.General.workDir = "."

        config.WorkQueueManager.logLevel = 'INFO'
        config.WorkQueueManager.pollInterval = 10
        config.WorkQueueManager.level = "GlobalQueue"
        return config

    def setupGlobalWorkqueue(self, **kwargs):
        """Return a workqueue instance"""
        kwargs.setdefault('rucioAccount', "wmcore_transferor")
        kwargs.setdefault('rucioAuthUrl', "https://cms-rucio-auth-int.cern.ch")
        kwargs.setdefault('rucioUrl', "http://cms-rucio-int.cern.ch")
        globalQ = globalQueue(DbName=self.globalQDB,
                              InboxDbName=self.globalQInboxDB,
                              QueueURL=self.globalQCouchUrl,
                              **kwargs)
        return globalQ

    def setupLocalQueue(self):
        """Create a local queue"""
        localQ = localQueue(DbName=self.localQDB,
                            InboxDbName=self.localQInboxDB,
                            QueueURL=self.localQCouchUrl,
                            Teams=["The A-Team", "some other bloke"],
                            ParentQueueCouchUrl=self.globalQCouchUrl,
                            CacheDir=self.testInit.testDir)
        return localQ

    @attr('integration')  # BROKEN
    def testReqMgrPollerAlgorithm(self):
        """ReqMgr reporting"""
        # don't actually talk to ReqMgr - mock it.
        globalQ = self.setupGlobalWorkqueue()
        localQ = self.setupLocalQueue()
        reqMgr = fakeReqMgr(splitter='Block')
        reqMgrInt = WorkQueueReqMgrInterface()
        reqMgrInt.reqMgr = reqMgr

        # 1st run should pull a request
        self.assertEqual(len(globalQ), 0)
        reqMgrInt(globalQ)
        self.assertEqual(len(globalQ), 2)
        self.assertEqual(reqMgr.status[reqMgr.names[0]], 'acquired')

        # local queue acquires and runs
        globalQ.updateLocationInfo()
        work = localQ.pullWork({'T2_XX_SiteA': 10000, 'T2_XX_SiteB': 10000})
        self.assertEqual(len(globalQ), 0)
        reqMgrInt(globalQ)
        self.assertEqual(reqMgr.status[reqMgr.names[0]], 'acquired')

        # start running work
        globalQ.setStatus('Running', WorkflowName=reqMgr.names[0])
        globalQ.performQueueCleanupActions()
        reqMgrInt(globalQ)  # report back to ReqMgr
        self.assertEqual(reqMgr.status[reqMgr.names[0]], 'running-closed')

        # finish work
        work = globalQ.status()
        globalQ.setStatus('Done', elementIDs=[x.id for x in work])
        reqMgrInt(globalQ)
        globalQ.performQueueCleanupActions()
        reqMgrInt(globalQ)
        self.assertEqual(reqMgr.status[reqMgr.names[0]], 'completed')
        # and removed from WorkQueue
        self.assertEqual(len(globalQ.status()), 0)

        # reqMgr problems should not crash client
        reqMgrInt.reqMgr = None
        reqMgrInt(globalQ)
        reqMgr._removeSpecs()

    @attr('integration')  # BROKEN
    def testReqMgrProgress(self):
        """ReqMgr interaction with block level splitting"""
        globalQ = self.setupGlobalWorkqueue()
        reqMgr = fakeReqMgr(splitter='Block')
        # reqMgrInt = WorkQueueReqMgrInterface(reqmgr2_endpoint=self.reqmgr2_endpoint)
        reqMgrInt = WorkQueueReqMgrInterface(reqmgr2_endpoint="https://fakeurl")
        reqMgrInt.reqMgr2 = reqMgr

        self.assertEqual(len(globalQ), 0)
        reqMgrInt(globalQ)
        self.assertEqual(len(globalQ), 2)
        globalQ.setStatus('Acquired', WorkflowName=reqMgr.names[0])
        reqMgrInt(globalQ)  # report back to ReqMgr
        self.assertEqual(reqMgr.status[reqMgr.names[0]], 'acquired')
        globalQ.setStatus('Running', WorkflowName=reqMgr.names[0])
        elements = globalQ.status()
        self.assertEqual(len(elements), 2)
        [globalQ.backend.updateElements(x.id, PercentComplete=75, PercentSuccess=25) for x in elements]
        elements = globalQ.status()
        globalQ.performQueueCleanupActions()
        reqMgrInt(globalQ)  # report back to ReqMgr
        self.assertEqual(reqMgr.status[reqMgr.names[0]], 'running-closed')
        globalQ.performQueueCleanupActions()
        reqMgrInt(globalQ)
        self.assertEqual(reqMgr.progress[reqMgr.names[0]]['percent_complete'],
                         75)
        self.assertEqual(reqMgr.progress[reqMgr.names[0]]['percent_success'],
                         25)
        self.assertEqual(reqMgr.status[reqMgr.names[0]], 'running-closed')
        globalQ.setStatus('Done', WorkflowName=reqMgr.names[0])
        reqMgrInt(globalQ)  # report back to ReqMgr
        globalQ.performQueueCleanupActions()
        reqMgrInt(globalQ)
        self.assertEqual(reqMgr.progress[reqMgr.names[0]]['percent_complete'],
                         75)
        self.assertEqual(reqMgr.status[reqMgr.names[0]], 'completed')
        reqMgr._removeSpecs()

    @attr('integration')  # BROKEN
    def testInvalidSpec(self):
        """Report invalid spec back to ReqMgr"""
        globalQ = self.setupGlobalWorkqueue()
        reqMgr = fakeReqMgr(inputDataset='thisdoesntexist')
        reqMgrInt = WorkQueueReqMgrInterface()
        reqMgrInt.reqMgr = reqMgr
        reqMgrInt(globalQ)
        self.assertEqual('failed', reqMgr.status[reqMgr.names[0]])
        self.assertTrue('No work in spec:' in reqMgr.msg[reqMgr.names[0]])
        reqMgr._removeSpecs()

        globalQ = self.setupGlobalWorkqueue()
        reqMgr = fakeReqMgr(dbsUrl='wrongprot://dbs.example.com')
        reqMgrInt = WorkQueueReqMgrInterface()
        reqMgrInt.reqMgr = reqMgr
        reqMgrInt(globalQ)
        self.assertEqual('failed', reqMgr.status[reqMgr.names[0]])
        self.assertTrue('DBS config error' in reqMgr.msg[reqMgr.names[0]])
        reqMgr._removeSpecs()

    @attr('integration')  # BROKEN
    def testCancelPickedUp(self):
        """WorkQueue cancels if canceled in ReqMgr"""
        globalQ = self.setupGlobalWorkqueue()
        reqMgr = fakeReqMgr()
        reqMgrInt = WorkQueueReqMgrInterface()
        reqMgrInt.reqMgr = reqMgr
        reqMgrInt(globalQ)
        # abort in reqmgr
        reqMgr.status[reqMgr.names[0]] = 'aborted'
        # workqueue will detect abort and cancel workflow
        reqMgrInt(globalQ)
        self.assertEqual(globalQ.status(WorkflowName=reqMgr.names[0])[0]['Status'], 'Canceled')
        # check stays canceled
        reqMgrInt(globalQ)
        self.assertEqual(globalQ.status(WorkflowName=reqMgr.names[0])[0]['Status'], 'Canceled')

    @attr('integration')  # BROKEN
    def testReqMgrWaitTime(self):
        """If running request finished in Reqmgr and no update locally for a long time decalre request done"""
        globalQ = self.setupGlobalWorkqueue()
        reqMgr = fakeReqMgr()
        reqMgrInt = WorkQueueReqMgrInterface()
        reqMgrInt.reqMgr = reqMgr
        reqMgrInt(globalQ)
        # set reqmgr status to done
        reqMgr.status[reqMgr.names[0]] = 'completed'
        # Only running elements are eligible for cleanup
        reqMgrInt(globalQ)
        self.assertEqual(globalQ.status(WorkflowName=reqMgr.names[0])[0]['Status'], 'Available')
        # Set running and element update time to the past (mimic time elapsing)
        globalQ.setStatus('Running', WorkflowName=reqMgr.names[0])
        element = globalQ.backend.inbox.document(reqMgr.names[0])
        element['updatetime'] = 0
        element['WMCore.WorkQueue.DataStructs.WorkQueueElement.WorkQueueElement']['Status'] = 'Running'
        globalQ.backend.inbox.commitOne(element)
        elements = [globalQ.backend.db.document(x.id) for x in globalQ.status(RequestName=reqMgr.names[0])]
        for element in elements:
            element['updatetime'] = 0
            element['WMCore.WorkQueue.DataStructs.WorkQueueElement.WorkQueueElement']['Status'] = 'Running'
            globalQ.backend.db.queue(element)
        globalQ.backend.db.commit()
        # workqueue should see an old done request and update status to match
        reqMgrInt(globalQ)
        self.assertEqual(globalQ.status(WorkflowName=reqMgr.names[0])[0]['Status'], 'Done')

    @attr('integration')  # BROKEN
    def testReqMgrOpenRequests(self):
        """Check the mechanics of open running requests"""
        # don't actually talk to ReqMgr - mock it.
        globalQ = self.setupGlobalWorkqueue()
        localQ = self.setupLocalQueue()
        reqMgr = fakeReqMgr(splitter='Block', openRunningTimeout=3600)
        reqMgrInt = WorkQueueReqMgrInterface()
        reqMgrInt.reqMgr = reqMgr

        # 1st run should pull a request
        self.assertEqual(len(globalQ), 0)
        reqMgrInt(globalQ)
        self.assertEqual(len(globalQ), 2)
        self.assertEqual(reqMgr.status[reqMgr.names[0]], 'acquired')

        # local queue acquires and runs
        globalQ.updateLocationInfo()
        localQ.pullWork({'T2_XX_SiteA': 10000, 'T2_XX_SiteB': 10000})
        self.assertEqual(len(globalQ), 0)
        reqMgrInt(globalQ)
        self.assertEqual(reqMgr.status[reqMgr.names[0]], 'acquired')

        # start running work
        globalQ.setStatus('Running', WorkflowName=reqMgr.names[0])
        globalQ.performQueueCleanupActions()
        reqMgrInt(globalQ)  # report back to ReqMgr
        self.assertEqual(reqMgr.status[reqMgr.names[0]], 'running-open')

        # Work should not be closed yet
        reqMgrInt(globalQ)
        globalQ.performQueueCleanupActions()
        self.assertEqual(reqMgr.status[reqMgr.names[0]], 'running-open')

        # Now add 2 new blocks to dbs, the reqMgr should put more work in the queue for the request
        GlobalParams.setNumOfBlocksPerDataset(GlobalParams.numOfBlocksPerDataset() + 2)
        reqMgrInt(globalQ)
        globalQ.performQueueCleanupActions()
        self.assertEqual(len(globalQ), 2)

        # Work that can be pulled normally and request stays in running-open
        globalQ.updateLocationInfo()
        localQ.pullWork({'T2_XX_SiteC': 10000})
        self.assertEqual(len(globalQ), 0)
        reqMgrInt(globalQ)
        self.assertEqual(reqMgr.status[reqMgr.names[0]], 'running-open')

        # Put the latest work to run
        globalQ.setStatus('Running', WorkflowName=reqMgr.names[0])
        globalQ.performQueueCleanupActions()
        reqMgrInt(globalQ)  # report back to ReqMgr
        self.assertEqual(reqMgr.status[reqMgr.names[0]], 'running-open')

        # Change the request status manually to close it
        reqMgr.status[reqMgr.names[0]] = 'running-closed'
        globalQ.performQueueCleanupActions()
        reqMgrInt(globalQ)  # report back to WorkQueue
        self.assertEqual(len(globalQ.backend.getInboxElements(OpenForNewData=False)), 1)

        # Put 1 more block in DBS for the dataset, request is closed so no more data is added
        GlobalParams.setNumOfBlocksPerDataset(GlobalParams.numOfBlocksPerDataset() + 1)
        reqMgrInt(globalQ)
        globalQ.performQueueCleanupActions()
        self.assertEqual(len(globalQ), 0)

        # finish work
        work = globalQ.status()
        globalQ.setStatus('Done', elementIDs=[x.id for x in work])
        reqMgrInt(globalQ)
        globalQ.performQueueCleanupActions()
        reqMgrInt(globalQ)
        self.assertEqual(reqMgr.status[reqMgr.names[0]], 'completed')
        # and removed from WorkQueue
        self.assertEqual(len(globalQ.status()), 0)

        # reqMgr problems should not crash client
        reqMgrInt.reqMgr = None
        reqMgrInt(globalQ)
        reqMgr._removeSpecs()


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
"""
_LoadFromName_

Oracle implementation of LoadFileset
"""

__all__ = []



from WMCore.WMBS.MySQL.Fileset.LoadFromName import LoadFromName as LoadFilesetMySQL

class LoadFromName(LoadFilesetMySQL):
    sql = LoadFilesetMySQL.sql



from __future__ import (division, print_function)

from future.utils import viewitems

from time import time
from WMCore.REST.HeartbeatMonitorBase import HeartbeatMonitorBase
from WMCore.WorkQueue.WorkQueue import globalQueue


class HeartbeatMonitor(HeartbeatMonitorBase):
    def __init__(self, rest, config):
        super(HeartbeatMonitor, self).__init__(rest, config)
        self.initialStatus = ['Available', 'Negotiating', 'Acquired']
        self.producer = "global_workqueue"
        self.docTypeAMQ = "cms_%s_info" % self.producer
        self.globalQ = globalQueue(logger=self.logger, **config.queueParams)

    def addAdditionalMonitorReport(self, config):
        """
        Collect some statistics for Global Workqueue and upload it to WMStats and
        MonIT.
        """
        tStart = time()
        self.logger.info("Collecting GlobalWorkqueue statistics...")

        # retrieve whole docs for these status in order to create site metrics
        results = self.globalQ.monitorWorkQueue(self.initialStatus)

        if self.postToAMQ:
            allDocs = self.buildMonITDocs(results)
            self.uploadToAMQ(allDocs)
        self.logger.info("%s executed in %.3f secs.", self.__class__.__name__, time() - tStart)

        return results

    def buildMonITDocs(self, stats):
        """
        Given the statistics that are uploaded to wmstats, create different
        documents to post to MonIT AMQ (aggregation-friendly docs).
        """
        mapMetricToType = {'uniqueJobsPerSite': 'work_site_unique',  # respecting data location constraints
                           'possibleJobsPerSite': 'work_site_possible',
                           'uniqueJobsPerSiteAAA': 'work_site_uniqueAAA',  # assume work can run anywhere
                           'possibleJobsPerSiteAAA': 'work_site_possibleAAA'}
        commonInfo = {"agent_url": "global_workqueue"}

        docs = []
        for status, data in viewitems(stats['workByStatus']):
            doc = dict()
            doc["type"] = "work_info"
            doc["status"] = status
            doc["num_elem"] = data.get('num_elem', 0)  # total number of workqueue elements
            doc["sum_jobs"] = data.get('sum_jobs', 0)  # total number of top level jobs
            doc["max_jobs_elem"] = data.get('max_jobs_elem', 0)  # largest # of jobs found in a WQE
            docs.append(doc)

        for status, data in viewitems(stats['workByStatusAndPriority']):
            for item in data:
                doc = dict()
                doc["type"] = "work_prio_status"
                doc["status"] = status
                doc.update(item)
                docs.append(doc)

        for item in stats['workByAgentAndStatus']:
            doc = dict()
            doc["type"] = "work_agent_status"
            doc["status"] = item['status']
            doc["agent_name"] = item['agent_name']
            doc["num_elem"] = item['num_elem']  # total number of workqueue elements
            doc["sum_jobs"] = item['sum_jobs']  # total number of top level jobs
            doc["max_jobs_elem"] = item['max_jobs_elem']  # largest # of jobs found in a WQE
            docs.append(doc)

        for item in stats['workByAgentAndPriority']:
            doc = dict()
            doc["type"] = "work_agent_prio"
            doc["priority"] = item['priority']
            doc["agent_name"] = item['agent_name']
            doc["num_elem"] = item['num_elem']  # total number of workqueue elements
            doc["sum_jobs"] = item['sum_jobs']  # total number of top level jobs
            doc["max_jobs_elem"] = item['max_jobs_elem']  # largest # of jobs found in a WQE
            docs.append(doc)

        # let's remap Jobs --> sum_jobs , and NumElems --> num_elem
        for metric in mapMetricToType:
            for status, sites in viewitems(stats[metric]):
                if not sites:
                    # no work in this status available for any sites, skip!
                    continue
                for site in sites:
                    doc = dict()
                    doc["type"] = mapMetricToType[metric]
                    doc["status"] = status
                    doc["site_name"] = site
                    doc['num_elem'] = sites[site]['num_elem']
                    doc['sum_jobs'] = int(sites[site]['sum_jobs'])
                    docs.append(doc)

        # mark every single document as being from global_workqueue
        for doc in docs:
            doc.update(commonInfo)

        self.logger.info("%i docs created to post to MonIT", len(docs))
        return docs



#!/usr/bin/env python
"""
__Controllers__

Controllers return java script and/or css from a static directory, after
minimising setting appropriate headers and etags and gzip.
"""




import os
from WMCore.WebTools.Page import TemplatedPage, exposejs, exposecss
from Controllers import Controllers
from cherrypy import expose

class Masthead(TemplatedPage, Controllers):
    """
    The WebTools masthead
    """
    def __init__(self, config):
        TemplatedPage.__init__(self, config)
        Controllers.__init__(self, config)

    @expose
    def index(self):
        return self.masthead()

    @exposejs
    def masthead (self):
        return self.templatepage ('masthead')

    @exposejs
    def masthead_table(self):
        return self.templatepage ('masthead_table')

    @exposecss
    def mastheadcss(self, site=None, *args, **kwargs):
        files = ['dmwt_main.css', 'dmwt_masthead.css']
        if site:
            files+= ['dmwt_masthead_%s.css' % site]
        path = __file__.rsplit('/',1)[0]
        data = ""
        for f in files:
            filename = "%s/css/%s" % (path, f)
            if os.path.exists(filename):
                lines = open(filename).readlines()
                for l in lines:
                    data = data + l

        return self.minify(self.templatepage ('data', {'data':data}))

    @exposecss
    def mastheadcss_table(self, site=None, *args, **kwargs):
        files = ['dmwt_main_table.css', 'dmwt_masthead.css']
        if site:
            files+= ['dmwt_masthead_table_%s.css' % site]
        path = __file__.rsplit('/',1)[0]
        data = ""
        for f in files:
            filename = "%s/css/%s" % (path, f)
            if os.path.exists(filename):
                lines = open(filename).readlines()
                for l in lines:
                    data = data + l

        return self.minify(self.templatePage ('data', {'data':data}))



#!/usr/bin/env python
"""
_Config_

Pseudo CMSSW config classes.
"""

class Container():
    """
    _Container_

    Empty class to emulate a PSet.
    """
    _type = None

    def setType(self, type):
        """
        _setType_

        Set the type of the object.
        """
        self._type = type
        return

    def type_(self):
        """
        _type__

        Retrieve the type of the object.
        """
        return self._type

class ConfigObject():
    """
    _ConfigObject_

    Emulate some sort of configuration parameter.
    """
    def __init__(self, type, value):
        self.type = type
        self.value = value
        return

    def setValue(self, value):
        self.value = value

    def __getitem__(self, item):
        if isinstance(self.value, list):
            return self.value[item]

class _Untracked(object):
    """
    _Untracked_

    Wrapper for untracker parameters.
    """
    @staticmethod
    def __call__(param):
        return param

    def __getattr__(self, name):
        class Factory(object):
            def __init__(self, name):
                self.name = name

            def __call__(self, *value, **params):
                param = globals()[self.name](*value, **params)
                return _Untracked.__call__(param)

        return Factory(name)

untracked = _Untracked()

def string(stringContents = None):
    """
    _string_

    String configuration parameter.
    """
    return ConfigObject("string", stringContents)

def vstring(stringContents = None):
    """
    _vstring_

    VString configuration parameter.
    """
    return ConfigObject("vstring", stringContents)

def uint32(value = 0):
    """
    _uint32_

    Unsigned 32bit integer parameter.
    """
    return ConfigObject("uint32", value)

def int32(value = 0):
    """
    _int32_

    Signed integer parameter.
    """
    return ConfigObject("uint32", value)

def bool(value = True):
    """
    _bool_

    Bool parameter.
    """
    return ConfigObject("bool", value)

def VLuminosityBlockRange():
    """
    _VLuminosityBlockRange_

    Array of luminosity block ranges.
    """
    return []

def PSet(**attributes):
    """
    _PSet_

    Create a new PSet with the given attributes.
    """
    newPSet = Container()
    for attributeName in attributes.keys():
        setattr(newPSet, attributeName, attributes[attributeName])

    return newPSet

def Service(serviceName, **kwargs):
    """
    _Service_

    Create a new service.
    """
    newService = Container()
    setattr(newService, "serviceName", serviceName)
    return newService



#!/usr/bin/env python
"""
_ListThresholdsForSubmit_

Oracle implementation of ResourceControl.ListThresholdsForSubmit
"""




from WMCore.ResourceControl.MySQL.ListThresholdsForSubmit \
     import ListThresholdsForSubmit as MySQLListThresholdsForSubmit

class ListThresholdsForSubmit(MySQLListThresholdsForSubmit):
    pass



#!/usr/bin/env python
"""
    WorkQueue.Policy.Start.Block tests
"""

from future.utils import viewitems, listvalues

import unittest

from WMCore_t.WMSpec_t.samples.MultiTaskProcessingWorkload import workload as MultiTaskProcessingWorkload
from WMCore_t.WorkQueue_t.WorkQueue_t import getFirstTask

from Utils.PythonVersion import PY3

from WMCore.DataStructs.LumiList import LumiList
from WMCore.Services.DBS.DBSErrors import DBSReaderError
from WMCore.Services.DBS.DBSReader import DBSReader
from WMCore.WMSpec.StdSpecs.ReReco import ReRecoWorkloadFactory
from WMCore.WorkQueue.Policy.Start.Block import Block
from WMCore.WorkQueue.WorkQueueExceptions import (WorkQueueWMSpecError, WorkQueueNoWorkError)
from WMQuality.Emulators.EmulatedUnitTestCase import EmulatedUnitTestCase
from WMQuality.Emulators.WMSpecGenerator.WMSpecGenerator import createConfig

rerecoArgs = ReRecoWorkloadFactory.getTestArguments()
rerecoArgs["SplittingAlgo"] = "LumiBased"
rerecoArgs["LumisPerJob"] = 8
rerecoArgs["ConfigCacheID"] = createConfig(rerecoArgs["CouchDBName"])

parentProcArgs = ReRecoWorkloadFactory.getTestArguments()
parentProcArgs.update(IncludeParents="True")
parentProcArgs["SplittingAlgo"] = "LumiBased"
parentProcArgs["LumisPerJob"] = 8


def rerecoWorkload(workloadName, arguments, assignArgs=None):
    factory = ReRecoWorkloadFactory()
    wmspec = factory.factoryWorkloadConstruction(workloadName, arguments)
    if assignArgs:
        args = factory.getAssignTestArguments()
        args.update(assignArgs)
        wmspec.updateArguments(args)
    return wmspec


class BlockTestCase(EmulatedUnitTestCase):
    splitArgs = dict(SliceType='NumberOfFiles', SliceSize=10)

    def setUp(self):
        super(BlockTestCase, self).setUp()
        if PY3:
            self.assertItemsEqual = self.assertCountEqual

    def tearDown(self):
        super(BlockTestCase, self).tearDown()

    def testTier1ReRecoWorkload(self):
        """Tier1 Re-reco workflow"""
        Tier1ReRecoWorkload = rerecoWorkload('ReRecoWorkload', rerecoArgs,
                                             assignArgs={'SiteWhitelist': ['T2_XX_SiteA']})
        Tier1ReRecoWorkload.data.request.priority = 69
        inputDataset = getFirstTask(Tier1ReRecoWorkload).inputDataset()
        dataset = "/%s/%s/%s" % (inputDataset.primary,
                                 inputDataset.processed,
                                 inputDataset.tier)
        dbs = {inputDataset.dbsurl: DBSReader(inputDataset.dbsurl)}
        for task in Tier1ReRecoWorkload.taskIterator():
            units, _, _ = Block(**self.splitArgs)(Tier1ReRecoWorkload, task)
            self.assertEqual(47, len(units))
            for unit in units:
                self.assertEqual(69, unit['Priority'])
                self.assertTrue(1 <= unit['Jobs'])
                self.assertEqual(Tier1ReRecoWorkload, unit['WMSpec'])
                self.assertEqual(task, unit['Task'])
                self.assertTrue(1 <= unit['NumberOfLumis'])
                self.assertTrue(1 <= unit['NumberOfFiles'])
                self.assertTrue(0 <= unit['NumberOfEvents'])
            self.assertEqual(len(units),
                             len(dbs[inputDataset.dbsurl].listFileBlocks(dataset)))

    def testMultiTaskProcessingWorkload(self):
        """Multi Task Processing Workflow"""
        datasets = []
        tasks, count = 0, 0
        for task in MultiTaskProcessingWorkload.taskIterator():
            tasks += 1
            inputDataset = task.inputDataset()
            datasets.append("/%s/%s/%s" % (inputDataset.primary,
                                           inputDataset.processed,
                                           inputDataset.tier))
        dbs = {inputDataset.dbsurl: DBSReader(inputDataset.dbsurl)}

        for task in MultiTaskProcessingWorkload.taskIterator():
            units, _, _ = Block(**self.splitArgs)(MultiTaskProcessingWorkload, task)
            self.assertEqual(58, len(units))

            for unit in units:
                self.assertTrue(1 <= unit['Jobs'])
                self.assertEqual(MultiTaskProcessingWorkload, unit['WMSpec'])
                self.assertEqual(task, unit['Task'])
            self.assertEqual(len(units),
                             len(dbs[inputDataset.dbsurl].listFileBlocks(datasets[0])))
            count += 1
        self.assertEqual(tasks, count)

    def testWhiteBlackLists(self):
        """Block/Run White/Black lists"""
        Tier1ReRecoWorkload = rerecoWorkload('ReRecoWorkload', rerecoArgs,
                                             assignArgs={'SiteWhitelist': ['T2_XX_SiteA']})
        inputDataset = getFirstTask(Tier1ReRecoWorkload).inputDataset()
        dataset = "/%s/%s/%s" % (inputDataset.primary, inputDataset.processed, inputDataset.tier)

        # No white/black lists
        newArgs = {}
        newArgs.update(rerecoArgs)
        workload = rerecoWorkload('ReRecoWorkload', newArgs,
                                  assignArgs={'SiteWhitelist': ['T2_XX_SiteA']})

        task = getFirstTask(workload)
        units, rejectedWork, badWork = Block(**self.splitArgs)(workload, task)
        self.assertEqual(len(units), 47)
        self.assertEqual(len(rejectedWork), 0)
        self.assertEqual(len(badWork), 0)

        # Block blacklist
        newArgs = {}
        newArgs.update(rerecoArgs)
        newArgs.update({'BlockBlacklist': [dataset + '#03fe83c2-0c23-11e1-b764-003048caaace']})
        workload = rerecoWorkload('ReRecoWorkload', newArgs,
                                  assignArgs={'SiteWhitelist': ['T2_XX_SiteA']})

        task = getFirstTask(workload)
        units, rejectedWork, badWork = Block(**self.splitArgs)(workload, task)
        self.assertEqual(len(units), 46)
        self.assertEqual(len(rejectedWork), 0)
        self.assertEqual(len(badWork), 0)
        self.assertNotEqual(list(units[0]['Inputs']), newArgs['BlockBlacklist'])

        # Block Whitelist
        newArgs = {}
        newArgs.update(rerecoArgs)
        newArgs.update({'BlockWhitelist': [dataset + '#03fe83c2-0c23-11e1-b764-003048caaace']})
        workload = rerecoWorkload('ReRecoWorkload', newArgs,
                                  assignArgs={'SiteWhitelist': ['T2_XX_SiteA']})

        task = getFirstTask(workload)
        units, rejectedWork, badWork = Block(**self.splitArgs)(workload, task)
        self.assertEqual(len(units), 1)
        self.assertEqual(len(rejectedWork), 0)
        self.assertEqual(len(badWork), 0)
        self.assertEqual(list(units[0]['Inputs']), newArgs['BlockWhitelist'])

        # Block Mixed Whitelist
        newArgs = {}
        newArgs.update(rerecoArgs)
        newArgs.update({'BlockWhitelist': [dataset + '#04be2fcc-0b8f-11e1-b764-003048caaace'],
                        'BlockBlacklist': [dataset + '#03fe83c2-0c23-11e1-b764-003048caaace']})
        workload = rerecoWorkload('ReRecoWorkload', newArgs,
                                  assignArgs={'SiteWhitelist': ['T2_XX_SiteA']})

        task = getFirstTask(workload)
        units, rejectedWork, badWork = Block(**self.splitArgs)(workload, task)
        self.assertEqual(len(units), 1)
        self.assertEqual(len(rejectedWork), 0)
        self.assertEqual(len(badWork), 0)
        self.assertEqual(list(units[0]['Inputs']), newArgs['BlockWhitelist'])

        # Run Whitelist
        newArgs = {}
        newArgs.update(rerecoArgs)
        newArgs.update({'RunWhitelist': [181367]})
        workload = rerecoWorkload('ReRecoWorkload', newArgs,
                                  assignArgs={'SiteWhitelist': ['T2_XX_SiteA']})

        task = getFirstTask(workload)
        units, rejectedWork, badWork = Block(**self.splitArgs)(workload, task)
        self.assertEqual(len(units), 1)
        self.assertEqual(len(rejectedWork), 46)
        self.assertEqual(len(badWork), 0)
        self.assertEqual(list(units[0]['Inputs']), [dataset + '#03fe83c2-0c23-11e1-b764-003048caaace'])

        # Run Blacklist
        newArgs = {}
        newArgs.update(rerecoArgs)
        newArgs.update({'RunBlacklist': [180899, 180992, 1]})
        workload = rerecoWorkload('ReRecoWorkload', newArgs,
                                  assignArgs={'SiteWhitelist': ['T2_XX_SiteA']})

        task = getFirstTask(workload)
        units, rejectedWork, badWork = Block(**self.splitArgs)(workload, task)
        self.assertEqual(len(units), 45)
        self.assertEqual(len(rejectedWork), 2)
        self.assertEqual(len(badWork), 0)
        self.assertEqual(list(units[0]['Inputs']), [dataset + '#217ea8d8-0c4f-11e1-b764-003048caaace'])

        # Run Mixed Whitelist
        newArgs = {}
        newArgs.update(rerecoArgs)
        newArgs.update({'RunBlacklist': [180899], 'RunWhitelist': [180992]})
        workload = rerecoWorkload('ReRecoWorkload', newArgs,
                                  assignArgs={'SiteWhitelist': ['T2_XX_SiteA']})

        task = getFirstTask(workload)
        units, rejectedWork, badWork = Block(**self.splitArgs)(workload, task)
        self.assertEqual(len(units), 1)
        self.assertEqual(len(rejectedWork), 46)
        self.assertEqual(len(badWork), 0)
        self.assertEqual(list(units[0]['Inputs']), [dataset + '#b469f816-0946-11e1-8347-003048caaace'])

    def testLumiMask(self):
        """Lumi mask test"""
        rerecoArgs2 = {}
        rerecoArgs2.update(rerecoArgs)
        rerecoArgs2["ConfigCacheID"] = createConfig(rerecoArgs2["CouchDBName"])
        dummyWorkload = rerecoWorkload('ReRecoWorkload', rerecoArgs2,
                                       assignArgs={'SiteWhitelist': ['T2_XX_SiteA']})

        # Block blacklist
        lumiWorkload = rerecoWorkload('ReRecoWorkload', rerecoArgs2,
                                      assignArgs={'SiteWhitelist': ['T2_XX_SiteA']})
        task = getFirstTask(lumiWorkload)
        # task.data.input.splitting.runs = ['1']
        task.data.input.splitting.runs = ['180992']
        task.data.input.splitting.lumis = ['1,1']
        units, rejectedWork, badWork = Block(**self.splitArgs)(lumiWorkload, task)
        self.assertEqual(len(units), 1)
        self.assertEqual(len(rejectedWork), 46)
        self.assertEqual(len(badWork), 0)

    def testDataDirectiveFromQueue(self):
        """Test data directive from queue"""
        Tier1ReRecoWorkload = rerecoWorkload('ReRecoWorkload', rerecoArgs,
                                             assignArgs={'SiteWhitelist': ['T2_XX_SiteA']})
        inputDataset = getFirstTask(Tier1ReRecoWorkload).inputDataset()
        dataset = "/%s/%s/%s" % (inputDataset.primary,
                                 inputDataset.processed,
                                 inputDataset.tier)
        dbs = {inputDataset.dbsurl: DBSReader(inputDataset.dbsurl)}
        for task in Tier1ReRecoWorkload.taskIterator():
            # Take dataset and force to run over only 1 block
            units, _, _ = Block(**self.splitArgs)(Tier1ReRecoWorkload, task,
                                                  {dataset + '#28315b28-0c5c-11e1-b764-003048caaace': []})
            self.assertEqual(1, len(units))
            for unit in units:
                self.assertEqual(1, unit['Jobs'])
                self.assertEqual(Tier1ReRecoWorkload, unit['WMSpec'])
                self.assertEqual(task, unit['Task'])
            self.assertNotEqual(len(units),
                                len(dbs[inputDataset.dbsurl].listFileBlocks(dataset)))

    def testLumiSplitTier1ReRecoWorkload(self):
        """Tier1 Re-reco workflow"""
        splitArgs = dict(SliceType='NumberOfLumis', SliceSize=1)

        Tier1ReRecoWorkload = rerecoWorkload('ReRecoWorkload', rerecoArgs,
                                             assignArgs={'SiteWhitelist': ['T2_XX_SiteA']})
        Tier1ReRecoWorkload.setStartPolicy('Block', **splitArgs)

        for task in Tier1ReRecoWorkload.taskIterator():
            units, rejectedWork, badWork = Block(**splitArgs)(Tier1ReRecoWorkload, task)
            self.assertEqual(47, len(units))
            for unit in units:
                self.assertTrue(1 <= unit['Jobs'])
            self.assertEqual(0, len(rejectedWork))
            self.assertEqual(0, len(badWork))

    def testRunWhitelist(self):
        """
        ReReco lumi split with Run whitelist
        This test may not do much of anything anymore since listRunLumis is not in DBS3
        """

        splitArgs = dict(SliceType='NumberOfLumis', SliceSize=1)

        Tier1ReRecoWorkload = rerecoWorkload('ReRecoWorkload', rerecoArgs,
                                             assignArgs={'SiteWhitelist': ['T2_XX_SiteA']})
        Tier1ReRecoWorkload.setStartPolicy('Block', **splitArgs)
        Tier1ReRecoWorkload.setRunWhitelist([180899, 180992])
        inputDataset = getFirstTask(Tier1ReRecoWorkload).inputDataset()

        dbs = {inputDataset.dbsurl: DBSReader(inputDataset.dbsurl)}
        for task in Tier1ReRecoWorkload.taskIterator():
            units, rejectedWork, badWork = Block(**splitArgs)(Tier1ReRecoWorkload, task)
            # Blocks 1 and 2 match run distribution
            self.assertEqual(2, len(units))
            self.assertEqual(len(rejectedWork), 45)
            self.assertEqual(len(badWork), 0)
            # Check number of jobs in element match number for
            # dataset in run whitelist
            wq_jobs = 0
            for unit in units:
                wq_jobs += unit['Jobs']
                # This fails. listRunLumis does not work correctly with DBS3,
                # returning None for the # of lumis in a run
                runLumis = dbs[inputDataset.dbsurl].listRunLumis(block=list(unit['Inputs'])[0])
                for run in runLumis:
                    if run in getFirstTask(Tier1ReRecoWorkload).inputRunWhitelist():
                        # This is what it is with DBS3 unless we calculate it
                        self.assertEqual(runLumis[run], None)
            self.assertEqual(2, int(wq_jobs))

    def testInvalidSpecs(self):
        """Specs with no work"""
        # no dataset
        processingSpec = rerecoWorkload('ReRecoWorkload', rerecoArgs,
                                        assignArgs={'SiteWhitelist': ['T2_XX_SiteA']})
        getFirstTask(processingSpec).data.input.dataset = None
        for task in processingSpec.taskIterator():
            self.assertRaises(WorkQueueWMSpecError, Block(), processingSpec, task)

        # invalid dbs url
        processingSpec = rerecoWorkload('ReRecoWorkload', rerecoArgs,
                                        assignArgs={'SiteWhitelist': ['T2_XX_SiteA']})
        getFirstTask(processingSpec).data.input.dataset.dbsurl = 'wrongprot://dbs.example.com'
        for task in processingSpec.taskIterator():
            self.assertRaises(DBSReaderError, Block(), processingSpec, task)

        # dataset non existent
        processingSpec = rerecoWorkload('ReRecoWorkload', rerecoArgs,
                                        assignArgs={'SiteWhitelist': ['T2_XX_SiteA']})
        getFirstTask(processingSpec).data.input.dataset.name = "/MinimumBias/FAKE-Filter-v1/RECO"
        for task in processingSpec.taskIterator():
            self.assertRaises(DBSReaderError, Block(), processingSpec, task)

        # invalid run whitelist
        processingSpec = rerecoWorkload('ReRecoWorkload', rerecoArgs,
                                        assignArgs={'SiteWhitelist': ['T2_XX_SiteA']})
        processingSpec.setRunWhitelist([666])  # not in this dataset
        for task in processingSpec.taskIterator():
            self.assertRaises(WorkQueueNoWorkError, Block(), processingSpec, task)

    def notestParentProcessing(self):
        # Does not work with a RAW dataset, need a different workload
        """
        test parent processing: should have the same results as rereco test
        with the parent flag and dataset.
        """
        parentProcArgs["ConfigCacheID"] = createConfig(parentProcArgs["CouchDBName"])
        factory = ReRecoWorkloadFactory()
        parentProcSpec = factory.factoryWorkloadConstruction('testParentProcessing', parentProcArgs)

        inputDataset = getFirstTask(parentProcSpec).inputDataset()
        dataset = "/%s/%s/%s" % (inputDataset.primary,
                                 inputDataset.processed,
                                 inputDataset.tier)
        dbs = {inputDataset.dbsurl: DBSReader(inputDataset.dbsurl)}
        for task in parentProcSpec.taskIterator():
            units, _, _ = Block(**self.splitArgs)(parentProcSpec, task)
            self.assertEqual(47, len(units))
            for unit in units:
                self.assertTrue(1 <= unit['Jobs'])
                self.assertEqual(parentProcSpec, unit['WMSpec'])
                self.assertEqual(task, unit['Task'])
                self.assertEqual(True, unit['ParentFlag'])
                self.assertEqual(1, len(unit['ParentData']))
            self.assertEqual(len(units),
                             len(dbs[inputDataset.dbsurl].listFileBlocks(dataset)))

    def testIgnore0SizeBlocks(self):
        """Ignore blocks with 0 files"""
        Tier1ReRecoWorkload = rerecoWorkload('ReRecoWorkload', rerecoArgs,
                                             assignArgs={'SiteWhitelist': ['T2_XX_SiteA']})
        Tier1ReRecoWorkload.setRunWhitelist([2, 3])

        for task in Tier1ReRecoWorkload.taskIterator():
            self.assertRaises(WorkQueueNoWorkError, Block(**self.splitArgs), Tier1ReRecoWorkload, task)

    def testContinuousSplittingSupport(self):
        """Can modify successfully policies for continuous splitting"""
        policyInstance = Block(**self.splitArgs)
        self.assertTrue(policyInstance.supportsWorkAddition(), "Block instance should support continuous splitting")
        Tier1ReRecoWorkload = rerecoWorkload('ReRecoWorkload', rerecoArgs,
                                             assignArgs={'SiteWhitelist': ['T2_XX_SiteA']})
        Tier1ReRecoWorkload.data.request.priority = 69
        inputDataset = getFirstTask(Tier1ReRecoWorkload).inputDataset()
        dataset = "/%s/%s/%s" % (inputDataset.primary,
                                 inputDataset.processed,
                                 inputDataset.tier)
        dbs = {inputDataset.dbsurl: DBSReader(inputDataset.dbsurl)}
        for task in Tier1ReRecoWorkload.taskIterator():
            units, _, _ = policyInstance(Tier1ReRecoWorkload, task)
            self.assertEqual(47, len(units))
            blocks = []  # fill with blocks as we get work units for them
            inputs = {}
            for unit in units:
                blocks.extend(list(unit['Inputs']))
                inputs.update(unit['Inputs'])
                self.assertEqual(69, unit['Priority'])
                self.assertTrue(1 <= unit['Jobs'])
                self.assertEqual(Tier1ReRecoWorkload, unit['WMSpec'])
                self.assertEqual(task, unit['Task'])
                self.assertTrue(1 <= unit['NumberOfLumis'])
                self.assertTrue(1 <= unit['NumberOfFiles'])
                self.assertTrue(0 <= unit['NumberOfEvents'])
            self.assertEqual(len(units),
                             len(dbs[inputDataset.dbsurl].listFileBlocks(dataset)))

        # Modify the spec and task, get first a fresh policy instance
        policyInstance = Block(**self.splitArgs)
        for task in Tier1ReRecoWorkload.taskIterator():
            policyInstance.modifyPolicyForWorkAddition({'ProcessedInputs': list(inputs)})
            self.assertRaises(WorkQueueNoWorkError, policyInstance, Tier1ReRecoWorkload, task)

        # Run one last time
        policyInstance = Block(**self.splitArgs)
        for task in Tier1ReRecoWorkload.taskIterator():
            policyInstance.modifyPolicyForWorkAddition({'ProcessedInputs': list(inputs)})
            self.assertRaises(WorkQueueNoWorkError, policyInstance, Tier1ReRecoWorkload, task)

        return

    def testDatasetLocation(self):
        """
        _testDatasetLocation_

        This is a function of all start policies so only test it here
        as there is no StartPolicyInterface unit test
        """
        policyInstance = Block(**self.splitArgs)
        # The policy instance must be called first to initialize the values
        Tier1ReRecoWorkload = rerecoWorkload('ReRecoWorkload', rerecoArgs,
                                             assignArgs={'SiteWhitelist': ['T2_XX_SiteA']})
        for task in Tier1ReRecoWorkload.taskIterator():
            policyInstance(Tier1ReRecoWorkload, task)
            outputs = policyInstance.getDatasetLocations(
                {'https://cmsweb-prod.cern.ch/dbs/prod/global/DBSReader': Tier1ReRecoWorkload.listInputDatasets()})
            for dataset in outputs:
                self.assertItemsEqual(outputs[dataset], ['T2_XX_SiteA', 'T2_XX_SiteB'])
        return

    def testPileupData(self):
        """
        _testPileupData_

        Check that every workqueue element split contains the pile up data
        if it is present in the workload.
        """
        for task in MultiTaskProcessingWorkload.taskIterator():
            units, _, _ = Block(**self.splitArgs)(MultiTaskProcessingWorkload, task)
            self.assertEqual(58, len(units))
            for unit in units:
                pileupData = unit["PileupData"]
                self.assertEqual(len(pileupData), 1)
                self.assertItemsEqual(listvalues(pileupData)[0], ['T2_XX_SiteA', 'T2_XX_SiteB', 'T2_XX_SiteC'])
        return

    def testWithMaskedBlocks(self):
        """
        _testWithMaskedBlocks_

        Test job splitting with masked blocks
        """

        Tier1ReRecoWorkload = rerecoWorkload('ReRecoWorkload', rerecoArgs,
                                             assignArgs={'SiteWhitelist': ['T2_XX_SiteA']})
        Tier1ReRecoWorkload.data.request.priority = 69
        task = getFirstTask(Tier1ReRecoWorkload)
        dummyDataset = task.inputDataset()

        task.data.input.splitting.runs = [181061, 180899]
        task.data.input.splitting.lumis = ['1,50,60,70', '1,1']
        lumiMask = LumiList(compactList={'206371': [[1, 50], [60, 70]], '180899': [[1, 1]], })

        units, _, _ = Block(**self.splitArgs)(Tier1ReRecoWorkload, task)

        nLumis = 0
        for unit in units:
            nLumis += unit['NumberOfLumis']

        self.assertEqual(len(lumiMask.getLumis()), nLumis)

    def testGetMaskedBlocks(self):
        """
        _testGetMaskedBlocks_

        Check that getMaskedBlocks is returning the correct information
        """

        rerecoArgs["ConfigCacheID"] = createConfig(rerecoArgs["CouchDBName"])
        factory = ReRecoWorkloadFactory()

        Tier1ReRecoWorkload = factory.factoryWorkloadConstruction('ReRecoWorkload', rerecoArgs)
        Tier1ReRecoWorkload.data.request.priority = 69
        task = getFirstTask(Tier1ReRecoWorkload)
        inputDataset = task.inputDataset()
        inputDataset.primary = 'SingleElectron'
        inputDataset.processed = 'StoreResults-Run2011A-WElectron-PromptSkim-v4-ALCARECO-NOLC-36cfce5a1d3f3ab4df5bd2aa0a4fa380'
        inputDataset.tier = 'USER'

        task.data.input.splitting.runs = [166921, 166429, 166911]
        task.data.input.splitting.lumis = ['40,70', '1,50', '1,5,16,20']
        lumiMask = LumiList(compactList={'166921': [[40, 70]], '166429': [[1, 50]], '166911': [[1, 5], [16, 20]], })
        inputLumis = LumiList(compactList={'166921': [[1, 67]], '166429': [[1, 91]], '166911': [[1, 104]], })
        dataset = "/%s/%s/%s" % (inputDataset.primary,
                                 inputDataset.processed,
                                 inputDataset.tier)
        dbs = DBSReader(inputDataset.dbsurl)
        maskedBlocks = Block(**self.splitArgs).getMaskedBlocks(task, dbs, dataset)
        for dummyBlock, files in viewitems(maskedBlocks):
            for dummyFile, lumiList in viewitems(files):
                self.assertEqual(str(lumiList), str(inputLumis & lumiMask))


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
"""
_SimpleMyProxy_t_
Test the two basic SimpleMyProxy operations.
There is NO need to source your UI before running these tests.
The SimpleMyProxy module has not ability to delegate/create proxy/myproxy so it assumes a myproxy already exists.
"""

import unittest
import os
import logging
import re

from nose.plugins.attrib import attr
from WMCore.Credential.SimpleMyProxy import SimpleMyProxy

MYPROXYSERVER = os.environ.get('MYPROXY_SERVER', 'myproxy.cern.ch')
MYPROXYPORT = 7512
SERVERDN = os.environ.get('SERVER_DN', '/DC=ch/DC=cern/OU=computers/CN=mattia-dev02.cern.ch')
SERVERKEY = os.environ.get('SERVER_KEY', '/data/certs/hostkey.pem')
SERVERCERT = os.environ.get('SERVER_CERT', '/data/certs/hostcert.pem')
USERKEY = os.environ.get('USER_KEY', os.path.join(os.environ.get('HOME'),'.globus/userkey.pem'))
USERCERT = os.environ.get('USER_CERT', os.path.join(os.environ.get('HOME'),'.globus/usercert.pem'))
USERNAME = "simple_myproxy_test"

REGCERT = re.compile(r"^[-]{5}BEGIN CERTIFICATE[-]{5}[a-zA-Z0-9\\\n \+-]+")

class SimpleMyProxyTest(unittest.TestCase):

    def setUp(self):
        """
        Setup for unit tests
        """
        logging.basicConfig(level=logging.DEBUG,
                            format='%(asctime)s %(name)-12s %(levelname)-8s %(message)s',
                            datefmt='%m-%d %H:%M',
                            filename='proxy_unittests.log',
                            filemode='w')
        logger_name = 'SimpleMyProxyTest'
        self.logger = logging.getLogger(logger_name)
        self.simplemyproxy = SimpleMyProxy({'logger': self.logger})

    def tearDown(self):
        """
        _tearDown_
        """
        return

    @attr("integration")
    def testgetMyProxyInfo(self):
        """
        Test if it's possible to get myproxy information
        """
        myproxyinfo = self.simplemyproxy.checkMyProxy(username=USERNAME, myproxyserver=MYPROXYSERVER, myproxyport=MYPROXYPORT,
                                                      keyfile=USERKEY, certfile=USERCERT)
        assert 'retriever' in myproxyinfo
        assert 'owner' in myproxyinfo
        assert 'start' in myproxyinfo
        assert 'end' in myproxyinfo
        assert myproxyinfo['end'] > myproxyinfo['start']

    @attr("integration")
    def testgetMyProxy(self):
        """
        Test if it is possible to retrieve a proxy from an existing myproxy
        """
        myproxy = self.simplemyproxy.logonRenewMyProxy(username=USERNAME, myproxyserver=MYPROXYSERVER, myproxyport=MYPROXYPORT,
                                                       keyfile=SERVERKEY, certfile=SERVERCERT, lifetime=43200)
        self.assertTrue(REGCERT.match(myproxy))


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
"""
_ThreadPool_t_

Unit tests for WorkerThreads.

"""
from __future__ import absolute_import
from __future__ import print_function

import logging
import threading
import time
import unittest

from Utils.Timers import timeFunction
# local import
from WMCore_t.WorkerThreads_t.Dummy import Dummy

from WMCore.WorkerThreads.BaseWorkerThread import BaseWorkerThread
from WMCore.WorkerThreads.WorkerThreadManager import WorkerThreadManager
from WMQuality.TestInit import TestInit


class DummyWorker1(BaseWorkerThread):
    """
    Dummy class to implement a minimal worker object, used to check that all
    required worker methods get called
    """

    def __init__(self):
        """
        Get a callback to the test object used to access "thread doing stuff"
        data
        """
        BaseWorkerThread.__init__(self)
        myThread = threading.currentThread()
        self.dummySetupCallback = myThread.dummySetupCallback
        self.dummyAlgoCallback = myThread.dummyAlgoCallback
        self.dummyTerminateCallback = myThread.dummyTerminateCallback
        if 'workerThreadManager' in dir(myThread):
            self.workerThreadManager = myThread.workerThreadManager

    def setup(self, parameters):
        """
        Check the worker setup method is called
        """
        logging.info("DummyWorker1 setup called")
        self.dummySetupCallback()

    @timeFunction
    def algorithm(self, parameters):
        """
        Check the algorithm method is called
        """
        logging.info("DummyWorker1 algorithm called")
        self.dummyAlgoCallback()

    def terminate(self, parameters):
        """
        Check the terminate method is called
        """
        logging.info("DummyWorker1 terminate called")
        self.dummyTerminateCallback()
        # FIXME why this call doesn't reach the parent class?!?!
        super(DummyWorker1, self).terminate(parameters)


class DummyWorker2(BaseWorkerThread):
    """
    A very basic dummy worker
    """

    @timeFunction
    def algorithm(self, parameters):
        pass


class ErrorWorker(DummyWorker1):
    """
    A worker that throws an error
    """

    @timeFunction
    def algorithm(self, parameters):
        # workerThreadManager will be added by Harness
        # that isnt used here so add manually
        myThread = threading.currentThread()
        myThread.workerThreadManager = self.workerThreadManager
        raise RuntimeError("ErrorWorker throws errors")


class WorkerThreadsTest(unittest.TestCase):
    """
    Unit tests for WorkerThreads
    """

    _setupCalled = False
    _algoCalled = False
    _terminateCalled = False

    def dummySetupCallback(self):
        """
        Callback for setup
        """
        WorkerThreadsTest._setupCalled = True

    def dummyAlgoCallback(self):
        """
        Callback for algo
        """
        WorkerThreadsTest._algoCalled = True

    def dummyTerminateCallback(self):
        """
        Callback for terminate
        """
        WorkerThreadsTest._terminateCalled = True

    def setUp(self):
        "make a logger instance and create tables"

        # initialization necessary for proper style.
        myThread = threading.currentThread()
        myThread.dialect = None
        myThread.transaction = None
        myThread.dbFactory = None

        self.testInit = TestInit(__file__)
        self.testInit.setLogging()
        self.testInit.setDatabaseConnection()

        # Attack check callbacks to current thread
        myThread = threading.currentThread()
        myThread.dummySetupCallback = self.dummySetupCallback
        myThread.dummyAlgoCallback = self.dummyAlgoCallback
        myThread.dummyTerminateCallback = self.dummyTerminateCallback

    def tearDown(self):
        """
        Deletion of database
        """
        # FIXME: this might not work if your not using socket.
        myThread = threading.currentThread()

        # Remove callback methods from thread
        myThread = threading.currentThread()
        del myThread.dummySetupCallback
        del myThread.dummyAlgoCallback
        del myThread.dummyTerminateCallback

    def testA(self):
        """
        Check worker methods get called. We sleep occasionally to ensure these
        asynchronous calls have enough time to be called
        """
        # Create a worker manager
        compDummy = Dummy()
        print('create manager')
        manager = WorkerThreadManager(compDummy)

        # Pause it
        print('pause workers')
        manager.pauseWorkers()

        # Add a worker, and check init method gets called
        print('add worker')
        manager.addWorker(DummyWorker1(), 1)
        time.sleep(3)
        self.assertEqual(WorkerThreadsTest._setupCalled, True)
        # Ensure the algo wasn't called whilst paused
        self.assertEqual(WorkerThreadsTest._algoCalled, False)

        print('resume workers')
        # Run the workers, pause, and check algo method gets called
        manager.resumeWorkers()
        time.sleep(3)
        manager.pauseWorkers()
        self.assertEqual(WorkerThreadsTest._algoCalled, True)

        print('terminate workers')
        # Terminate the workers, and check terminate method gets called
        manager.terminateWorkers()
        time.sleep(3)
        self.assertEqual(WorkerThreadsTest._terminateCalled, True)

    def testB(self):
        """
        Check we can terminate before pausing workers
        """
        compDummy = Dummy()
        print('create manager')
        manager = WorkerThreadManager(compDummy)
        print('add worker')
        manager.addWorker(DummyWorker2(), 1)
        print('terminate worker')
        manager.terminateWorkers()

    def testWorkerError(self):
        """If a worker raises an exception terminate the entire component"""
        compDummy = Dummy()

        print('create manager')
        manager = WorkerThreadManager(compDummy)
        # needed for handling errors - harness would generally set this
        myThread = threading.currentThread()
        myThread.workerThreadManager = manager

        print('add workers')
        manager.addWorker(DummyWorker1(), 0.1)
        manager.addWorker(ErrorWorker(), 0.1)

        print('run workers, one will throw an error')
        manager.resumeWorkers()

        # should do something smarter here
        # too short a time and threads havent exited yet
        time.sleep(6)

        # all threads should have ended after worker raised exception
        self.assertEqual(manager.activeThreadCount, 0)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
"""
_WMInit

Init class that can be used by external projects
that only use part of the libraries
"""
from __future__ import print_function

import logging
import os
import os.path
import sys
import threading
import traceback

from WMCore.Configuration import loadConfigurationFile
from WMCore.DAOFactory import DAOFactory
from WMCore.Database.DBFactory import DBFactory
from WMCore.Database.Transaction import Transaction
from WMCore.WMBase import getWMBASE
from WMCore.WMException import WMException
from WMCore.WMFactory import WMFactory


class WMInitException(WMException):
    """
    WMInitException

    You should never, ever see one of these.
    I'm not optimistic that this will be the case.
    """


def connectToDB():
    """
    _connectToDB_

    Connect to the database specified in the WMAgent config.
    """
    if "WMAGENT_CONFIG" not in os.environ:
        print("Please set WMAGENT_CONFIG to point at your WMAgent configuration.")
        sys.exit(1)

    if not os.path.exists(os.environ["WMAGENT_CONFIG"]):
        print("Can't find config: %s" % os.environ["WMAGENT_CONFIG"])
        sys.exit(1)

    wmAgentConfig = loadConfigurationFile(os.environ["WMAGENT_CONFIG"])

    if not hasattr(wmAgentConfig, "CoreDatabase"):
        print("Your config is missing the CoreDatabase section.")
        sys.exit(1)

    socketLoc = getattr(wmAgentConfig.CoreDatabase, "socket", None)
    connectUrl = getattr(wmAgentConfig.CoreDatabase, "connectUrl", None)
    (dialect, junk) = connectUrl.split(":", 1)

    myWMInit = WMInit()
    myWMInit.setDatabaseConnection(dbConfig=connectUrl, dialect=dialect,
                                   socketLoc=socketLoc)
    return


class WMInit(object):
    def __init__(self):
        return

    def getWMBASE(self):
        """ for those that don't want to use the static version"""
        return getWMBASE()

    def setLogging(self, logFile=None, logName=None, logLevel=logging.INFO, logExists=True):
        """
        Sets logging parameters, depending on the settings,
        this method will create a logging file.
        """
        # use logName as name for file is no log file is given
        if not logExists:
            logging.basicConfig(level=logLevel, \
                                format='%(asctime)s %(name)-12s %(levelname)-8s %(message)s', \
                                datefmt='%m-%d %H:%M', \
                                filename='%s.log' % logFile, \
                                filemode='w')
            logging.debug("Log file ready")

        myThread = threading.currentThread()
        if logName != None:
            myThread.logger = logging.getLogger(logName)
        else:
            myThread.logger = logging.getLogger()

    def setDatabaseConnection(self, dbConfig, dialect, socketLoc=None):
        """
        Sets the default connection parameters, without having to worry
        much on what attributes need to be set. This is esepcially
        advantagous for developers of third party projects that want
        to use only parts of the WMCore lib.

        The class differentiates between different formats used by external
        projects. External project formats that are supported can activated
        it by setting the flavor flag.
        """
        myThread = threading.currentThread()
        if getattr(myThread, "dialect", None) != None:
            # Database is already initialized, we'll create a new
            # transaction and move on.
            if hasattr(myThread, "transaction"):
                if myThread.transaction != None:
                    myThread.transaction.commit()

            myThread.transaction = Transaction(myThread.dbi)
            return

        options = {}
        if dialect.lower() == 'mysql':
            dialect = 'MySQL'
            if socketLoc != None:
                options['unix_socket'] = socketLoc
        elif dialect.lower() == 'oracle':
            dialect = 'Oracle'
        elif dialect.lower() == 'http':
            dialect = 'CouchDB'
        else:
            msg = "Unsupported dialect %s !" % dialect
            logging.error(msg)
            raise WMInitException(msg)

        myThread.dialect = dialect
        myThread.logger = logging
        myThread.dbFactory = DBFactory(logging, dbConfig, options)
        myThread.dbi = myThread.dbFactory.connect()

        # The transaction object will begin a transaction as soon as it is
        # initialized.  I'd rather have the user handle that, so we'll commit
        # it here.
        myThread.transaction = Transaction(myThread.dbi)
        myThread.transaction.commit()

        return

    def setSchema(self, modules=None, params=None):
        """
        Creates the schema in the database based on the modules
        input.

        This method needs to have been preceded by the
        setDatabaseConnection.
        """
        modules = modules or []
        myThread = threading.currentThread()

        parameters = None
        flag = False
        # Set up for typical DBCreator format: logger, dbi, params
        if params != None:
            parameters = [None, None, params]
            flag = True

        myThread.transaction.begin()
        for factoryName in modules:
            # need to create these tables for testing.
            # notice the default structure: <dialect>/Create
            factory = WMFactory(factoryName, factoryName + "." + myThread.dialect)

            create = factory.loadObject("Create", args=parameters, listFlag=flag)
            createworked = create.execute(conn=myThread.transaction.conn,
                                          transaction=myThread.transaction)
            if createworked:
                logging.debug("Tables for " + factoryName + " created")
            else:
                logging.debug("Tables " + factoryName + " could not be created.")
        myThread.transaction.commit()

    def clearDatabase(self, modules=None):
        """
        Database deletion. Global, ignore modules.
        """
        myThread = threading.currentThread()
        if hasattr(myThread, 'transaction') and getattr(myThread.transaction, 'transaction', None):
            # Then we have an open transaction
            # We should try and close it first
            try:
                myThread.transaction.commit()
            except:
                try:
                    myThread.transaction.rollback()
                except:
                    pass

        # Setup the DAO
        daoFactory = DAOFactory(package="WMCore.Database",
                                logger=myThread.logger,
                                dbinterface=myThread.dbi)
        destroyDAO = daoFactory(classname="Destroy")

        # Actually run a transaction and delete the DB
        try:
            destroyDAO.execute()
        except Exception as ex:
            msg = "Critical error while attempting to delete entire DB!\n"
            msg += str(ex)
            msg += str(traceback.format_exc())
            logging.error(msg)
            raise WMInitException(msg)

        return

    def checkDatabaseContents(self):
        """
        _checkDatabaseContents_

        Check and see if anything is in the database.
        This should be called by methods about to build the schema to make sure
        that the DB itself is empty.
        """

        myThread = threading.currentThread()
        daoFactory = DAOFactory(package="WMCore.Database",
                                logger=myThread.logger,
                                dbinterface=myThread.dbi)

        testDAO = daoFactory(classname="ListUserContent")

        result = testDAO.execute()
        myThread.dbi.engine.dispose()

        return result



#!/usr/bin/env python
# pylint: disable=E1101,C0103,R0902
"""
Component test TestComponent module and the harness
"""
from __future__ import print_function

import os
import threading
import time
import unittest

import nose

from WMCore.Agent.Daemon.Details import Details
from WMCore.Database.Transaction import Transaction
from WMCore.WMFactory import WMFactory
from WMCore_t.Agent_t.TestComponent import TestComponent
from WMQuality.TestInit import TestInit


class HarnessTest(unittest.TestCase):
    """
    TestCase for TestComponent module
    """
    tempDir = None

    def setUp(self):
        """
        setup for test.
        """
        self.testInit = TestInit(__file__)
        self.testInit.setLogging()
        self.testInit.setDatabaseConnection()
        self.testInit.setSchema()

    def tearDown(self):
        """
        Delete database
        """
        self.testInit.clearDatabase()

    def testB(self):
        raise nose.SkipTest
        config = self.testInit.getConfiguration()
        self.tempDir = self.testInit.generateWorkDir(config)
        config.component_("TestComponent")
        config.TestComponent.logLevel = 'INFO'
        config.section_("General")
        config.TestComponent.componentDir = os.path.join( \
            self.tempDir, "Components/TestComponent1")
        config.General.workDir = config.TestComponent.componentDir

        os.makedirs(config.TestComponent.componentDir)
        # as this is a test we build the string from our global environment
        # parameters normally you put this straight into the DefaultConfig.py file:
        # testInit.getConfiguration returns from the environment variable by default
        testComponent = TestComponent(config)
        testComponent.prepareToStart()

        testComponent.handleMessage('LogState', '')
        testComponent.handleMessage('TestMessage1', 'TestMessag1Payload')
        testComponent.handleMessage('TestMessage2', 'TestMessag2Payload')
        testComponent.handleMessage('TestMessage3', 'TestMessag3Payload')
        testComponent.handleMessage('TestMessage4', 'TestMessag4Payload')
        testComponent.handleMessage('Logging.DEBUG', '')
        testComponent.handleMessage('Logging.WARNING', '')
        testComponent.handleMessage('Logging.CRITICAL', '')
        testComponent.handleMessage('Logging.ERROR', '')
        testComponent.handleMessage('Logging.INFO', '')
        testComponent.handleMessage('Logging.SQLDEBUG', '')
        testComponent.handleMessage('TestComponent:Logging.DEBUG', '')
        testComponent.handleMessage('TestComponent:Logging.WARNING', '')
        testComponent.handleMessage('TestComponent:Logging.CRITICAL', '')
        testComponent.handleMessage('TestComponent:Logging.ERROR', '')
        testComponent.handleMessage('TestComponent:Logging.INFO', '')
        testComponent.handleMessage('TestComponent:Logging.SQLDEBUG', '')
        # test a non existing message (to generate an error)
        errorMsg = ''
        try:
            testComponent.handleMessage('NonExistingMessageType', '')
        except Exception as ex:
            errorMsg = str(ex)
        self.assertTrue(errorMsg.startswith('Message NonExistingMessageType with payload'))

    def testC(self):
        raise nose.SkipTest
        config = self.testInit.getConfiguration()
        self.tempDir = self.testInit.generateWorkDir(config)
        config.component_("TestComponent")
        config.TestComponent.logLevel = 'INFO'
        config.section_("General")
        # try starting a component as a daemon:
        config.TestComponent.componentDir = os.path.join( \
            self.tempDir, "Components/TestComponent1")
        os.makedirs(config.TestComponent.componentDir)
        testComponent = TestComponent(config)
        # we set the parent to true as we are testing
        testComponent.startDaemon(keepParent=True)
        print('trying to kill the component')
        time.sleep(2)
        daemonFile = os.path.join(config.TestComponent.componentDir, "Daemon.xml")
        details = Details(daemonFile)
        print('Is component alive: ' + str(details.isAlive()))
        time.sleep(2)
        details.killWithPrejudice()
        print('Daemon killed')

    def testD(self):
        raise nose.SkipTest
        config = self.testInit.getConfiguration()
        config.component_("TestComponent")
        config.TestComponent.logLevel = 'INFO'
        config.section_("General")
        self.tempDir = self.testInit.generateWorkDir(config)
        # try starting a component as a daemon:
        config.TestComponent.componentDir = os.path.join( \
            self.tempDir, "Components/TestComponent2")
        os.makedirs(config.TestComponent.componentDir)
        testComponent = TestComponent(config)
        # we set the parent to true as we are testing
        testComponent.startDaemon(keepParent=True)
        time.sleep(2)
        daemonFile = os.path.join(config.TestComponent.componentDir, "Daemon.xml")
        details = Details(daemonFile)
        print('Is component alive: ' + str(details.isAlive()))

        # create msgService to send stop message.
        myThread = threading.currentThread()
        factory = WMFactory("msgService", "WMCore.MsgService." + \
                            myThread.dialect)
        myThread.transaction = Transaction(myThread.dbi)
        msgService = factory.loadObject("MsgService")
        msgService.registerAs("HarnessTest")
        myThread.transaction.commit()

        print('Publish a stop message to test if the component shutsdown gracefully')
        myThread.transaction.begin()
        msg = {'name': 'Stop', 'payload': ''}
        msgService.publish(msg)
        myThread.transaction.commit()

        msgService.finish()

        while details.isAlive():
            print('Component has not received stop message')
            time.sleep(2)
        print('Daemon shutdown gracefully')


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# -*- coding: ISO-8859-1 -*-
# pylint: disable=R0913,W0702,R0914,R0912,R0201
"""
File: pycurl_manager.py
Author: Valentin Kuznetsov <vkuznet@gmail.com>
Description: a basic wrapper around pycurl library.
The RequestHandler class provides basic APIs to get data
from a single resource or submit mutliple requests to
underlying data-services.

Examples:
# CERN SSO: http://linux.web.cern.ch/linux/docs/cernssocookie.shtml
# use RequestHandler with CERN SSO enabled site
mgr = RequestHandler()
url = "https://cms-gwmsmon.cern.ch/prodview/json/site_summary"
params = {}
tfile = tempfile.NamedTemporaryFile()
cern_sso_cookie(url, tfile.name, cert, ckey)
cookie = {url: tfile.name}
header, data = mgr.request(url3, params, cookie=cookie)
if header.status != 200:
    print "ERROR"

# fetch multiple urls at onces from various urls
tfile = tempfile.NamedTemporaryFile()
ckey = os.path.join(os.environ['HOME'], '.globus/userkey.pem')
cert = os.path.join(os.environ['HOME'], '.globus/usercert.pem')
url1 = "https://cmsweb-prod.cern.ch/dbs/prod/global/DBSReader/help"
url2 = "https://cmsweb-prod.cern.ch/dbs/prod/global/DBSReader/datatiers"
url3 = "https://cms-gwmsmon.cern.ch/prodview/json/site_summary"
cern_sso_cookie(url3, tfile.name, cert, ckey)
cookie = {url3: tfile.name}
urls = [url1, url2, url3]
data = getdata(urls, ckey, cert, cookie=cookie)
for row in data:
    print(row)
"""
from __future__ import print_function
from future import standard_library
standard_library.install_aliases()

from builtins import str, range, object
from past.builtins import basestring
from future.utils import viewitems


# system modules
import copy
import json
import gzip
import logging
import os
import re
import subprocess
import pycurl
from io import BytesIO
import http.client
from urllib.parse import urlencode

from Utils.Utilities import encodeUnicodeToBytes, decodeBytesToUnicode
from Utils.PortForward import portForward, PortForward
from Utils.TokenManager import TokenManager


def decompress(body, headers):
    """
    Helper function to decompress given body if HTTP headers contains gzip encoding
    :param body: bytes
    :param headers: dict
    :return: decode body
    """
    encoding = ""
    for header, value in headers.items():
        if header.lower() == 'content-encoding' and 'gzip' in value.lower():
            encoding = 'gzip'
            break
    if encoding != 'gzip':
        return body

    try:
        return gzip.decompress(body)
    except Exception as exc:
        logger = logging.getLogger()
        msg = "While processing decompress function with headers: %s, " % headers
        msg += "we were unable to decompress gzip content. Details: %s. " % str(exc)
        msg += "Considering response body as uncompressed."
        logger.exception(msg)
        return body

class ResponseHeader(object):
    """ResponseHeader parses HTTP response header"""

    def __init__(self, response):
        super(ResponseHeader, self).__init__()
        self.header = {}
        self.reason = ''
        self.fromcache = False
        self.parse(response)

    def parse(self, response):
        """Parse response header and assign class member data"""
        startRegex = r"^HTTP/\d.\d \d{3}"
        continueRegex = r"^HTTP/\d.\d 100"  # Continue: client should continue its request
        replaceRegex = r"^HTTP/\d.\d"

        response = decodeBytesToUnicode(response)

        for row in response.split('\r'):
            row = row.replace('\n', '')
            if not row:
                continue
            if re.search(startRegex, row):
                if re.search(continueRegex, row):
                    continue
                res = re.sub(replaceRegex, "", row).strip()
                status, reason = res.split(' ', 1)
                self.status = int(status)
                self.reason = reason
                continue
            try:
                key, val = row.split(':', 1)
                self.header[key.strip()] = val.strip()
            except:
                pass

    def getReason(self):
        """
        Return the HTTP request reason
        """
        return self.reason

    def getHeader(self):
        """
        Return the header dictionary object
        """
        return self.header

    def getHeaderKey(self, keyName):
        """
        Provided a key name, return it from the HTTP header.
        Note that - by design - header keys are meant to be
        case insensitive
        :param keyName: a header key name to be looked up
        :return: the value for that header key, or None if not found
        """
        for keyHea, valHea in self.header.items():
            if keyHea.lower() == keyName.lower():
                return valHea


class RequestHandler(object):
    """
    RequestHandler provides APIs to fetch single/multiple
    URL requests based on pycurl library
    """

    def __init__(self, config=None, logger=None):
        super(RequestHandler, self).__init__()
        if not config:
            config = {}
        defaultOpts = pycurl_options()
        self.nosignal = config.get('nosignal', defaultOpts['NOSIGNAL'])
        self.timeout = config.get('timeout', defaultOpts['TIMEOUT'])
        self.connecttimeout = config.get('connecttimeout', defaultOpts['CONNECTTIMEOUT'])
        self.followlocation = config.get('followlocation', defaultOpts['FOLLOWLOCATION'])
        self.maxredirs = config.get('maxredirs', defaultOpts['MAXREDIRS'])
        self.logger = logger if logger else logging.getLogger()
        self.tokenLocation = config.get('iam_token_file', '')
        if self.tokenLocation:
            self.tmgr = TokenManager(self.tokenLocation)
        else:
            self.tmgr = None

    def encode_params(self, params, verb, doseq, encode):
        """ Encode request parameters for usage with the 4 verbs.
            Assume params is already encoded if it is a string and
            uses a different encoding depending on the HTTP verb
            (either json.dumps or urllib.urlencode)
        """
        if not encode:
            return params
        # data is already encoded, just return it
        if isinstance(params, basestring):
            return params

        # data is not encoded, we need to do that
        if verb in ['GET', 'HEAD']:
            if params:
                encoded_data = urlencode(params, doseq=doseq)
            else:
                return ''
        else:
            if params:
                encoded_data = json.dumps(params)
            else:
                return {}

        return encoded_data

    def set_opts(self, curl, url, params, headers,
                 ckey=None, cert=None, capath=None, verbose=None,
                 verb='GET', doseq=True, encode=False, cainfo=None, cookie=None):
        """Set options for given curl object, params should be a dictionary"""
        if not (isinstance(params, (dict, basestring)) or params is None):
            raise TypeError("pycurl parameters should be passed as dictionary or an (encoded) string")
        #  ensure the original headers object remains unchanged
        headers = headers or {}  # if it's None, then make it a dict
        thisHeaders = copy.deepcopy(headers)
        curl.setopt(pycurl.NOSIGNAL, self.nosignal)
        curl.setopt(pycurl.TIMEOUT, self.timeout)
        curl.setopt(pycurl.CONNECTTIMEOUT, self.connecttimeout)
        curl.setopt(pycurl.FOLLOWLOCATION, self.followlocation)
        curl.setopt(pycurl.MAXREDIRS, self.maxredirs)

        if cookie and url in cookie:
            curl.setopt(pycurl.COOKIEFILE, cookie[url])
            curl.setopt(pycurl.COOKIEJAR, cookie[url])

        encoded_data = self.encode_params(params, verb, doseq, encode)

        if verb == 'GET':
            if encoded_data:
                url = url + '?' + encoded_data
        elif verb == 'HEAD':
            if encoded_data:
                url = url + '?' + encoded_data
            curl.setopt(pycurl.CUSTOMREQUEST, verb)
            curl.setopt(pycurl.HEADER, 1)
            curl.setopt(pycurl.NOBODY, True)
        elif verb == 'POST':
            curl.setopt(pycurl.POST, 1)
            if encoded_data:
                curl.setopt(pycurl.POSTFIELDS, encoded_data)
        elif verb == 'DELETE' or verb == 'PUT':
            curl.setopt(pycurl.CUSTOMREQUEST, verb)
            curl.setopt(pycurl.HTTPHEADER, ['Transfer-Encoding: chunked'])
            if encoded_data:
                curl.setopt(pycurl.POSTFIELDS, encoded_data)
        else:
            raise Exception('Unsupported HTTP method "%s"' % verb)

        if self.tmgr:
            token = self.tmgr.getToken()
            if token:
                thisHeaders['Authorization'] = 'Bearer {}'.format(token)

        if verb in ('POST', 'PUT'):
            # only these methods (and PATCH) require this header
            thisHeaders["Content-Length"] = str(len(encoded_data))

        # we must pass url as a bytes data-type, otherwise pycurl will fail with error
        # TypeError: invalid arguments to setopt
        # see https://curl.haxx.se/mail/curlpython-2007-07/0001.html
        curl.setopt(pycurl.URL, encodeUnicodeToBytes(url))
        # In order to enable service intercommunication with compressed HTTP body,
        # we need to enable this header here, in case it has not been provided by upstream.
        thisHeaders.setdefault("Accept-Encoding", "gzip")
        curl.setopt(pycurl.HTTPHEADER, [encodeUnicodeToBytes("%s: %s" % (k, v)) for k, v in viewitems(thisHeaders)])

        bbuf = BytesIO()
        hbuf = BytesIO()
        curl.setopt(pycurl.WRITEFUNCTION, bbuf.write)
        curl.setopt(pycurl.HEADERFUNCTION, hbuf.write)
        if capath:
            curl.setopt(pycurl.CAPATH, capath)
            curl.setopt(pycurl.SSL_VERIFYPEER, True)
            if cainfo:
                curl.setopt(pycurl.CAINFO, cainfo)
        else:
            curl.setopt(pycurl.SSL_VERIFYPEER, False)
        if ckey:
            curl.setopt(pycurl.SSLKEY, ckey)
        if cert:
            curl.setopt(pycurl.SSLCERT, cert)
        if verbose:
            curl.setopt(pycurl.VERBOSE, True)
            curl.setopt(pycurl.DEBUGFUNCTION, self.debug)
        return bbuf, hbuf

    def debug(self, debug_type, debug_msg):
        """Debug callback implementation"""
        print("debug(%d): %s" % (debug_type, debug_msg))

    def parse_body(self, data, decode=False):
        """
        Parse body part of URL request (by default use json).
        This method can be overwritten.
        """
        if decode:
            try:
                res = json.loads(data)
                return res
            except ValueError as exc:
                msg = 'Unable to load JSON data, %s, data type=%s, pass as is' \
                      % (str(exc), type(data))
                logging.debug(msg)
                return data
        else:
            return data

    def parse_header(self, header):
        """
        Parse response header.
        This method can be overwritten.
        """
        return ResponseHeader(header)

    @portForward(8443)
    def request(self, url, params, headers=None, verb='GET',
                verbose=0, ckey=None, cert=None, capath=None,
                doseq=True, encode=False, decode=False, cainfo=None, cookie=None):
        """Fetch data for given set of parameters"""
        curl = pycurl.Curl()
        bbuf, hbuf = self.set_opts(curl, url, params, headers, ckey, cert, capath,
                                   verbose, verb, doseq, encode, cainfo, cookie)
        curl.perform()
        if verbose:
            print(verb, url, params, headers)
        header = self.parse_header(hbuf.getvalue())
        data = bbuf.getvalue()
        data = decompress(data, header.header)
        if header.status < 300:
            if verb == 'HEAD':
                data = ''
            else:
                data = self.parse_body(data, decode)
        else:
            msg = 'url=%s, code=%s, reason=%s, headers=%s, result=%s' \
                  % (url, header.status, header.reason, header.header, data)
            exc = http.client.HTTPException(msg)
            setattr(exc, 'req_data', params)
            setattr(exc, 'req_headers', headers)
            setattr(exc, 'url', url)
            setattr(exc, 'result', data)
            setattr(exc, 'status', header.status)
            setattr(exc, 'reason', header.reason)
            setattr(exc, 'headers', header.header)
            bbuf.flush()
            hbuf.flush()
            raise exc

        bbuf.flush()
        hbuf.flush()
        return header, data

    def getdata(self, url, params, headers=None, verb='GET',
                verbose=0, ckey=None, cert=None, doseq=True,
                encode=False, decode=False, cookie=None):
        """Fetch data for given set of parameters"""
        _, data = self.request(url=url, params=params, headers=headers, verb=verb,
                               verbose=verbose, ckey=ckey, cert=cert, doseq=doseq,
                               encode=encode, decode=decode, cookie=cookie)
        return data

    def getheader(self, url, params, headers=None, verb='GET',
                  verbose=0, ckey=None, cert=None, doseq=True):
        """Fetch HTTP header"""
        header, _ = self.request(url, params, headers, verb,
                                 verbose, ckey, cert, doseq=doseq)
        return header

    @portForward(8443)
    def multirequest(self, url, parray, headers=None,
                     ckey=None, cert=None, verbose=None, cookie=None):
        """Fetch data for given set of parameters"""
        multi = pycurl.CurlMulti()
        for params in parray:
            curl = pycurl.Curl()
            bbuf, hbuf = \
                self.set_opts(curl, url, params, headers, ckey=ckey, cert=cert,
                              verbose=verbose, cookie=cookie)
            multi.add_handle(curl)
            while True:
                ret, num_handles = multi.perform()
                if ret != pycurl.E_CALL_MULTI_PERFORM:
                    break
            while num_handles:
                ret = multi.select(1.0)
                if ret == -1:
                    continue
                while True:
                    ret, num_handles = multi.perform()
                    if ret != pycurl.E_CALL_MULTI_PERFORM:
                        break
            dummyNumq, response, dummyErr = multi.info_read()
            for _respItem in response:
                data = decodeBytesToUnicode(bbuf.getvalue())
                data = json.loads(data)
                if isinstance(data, dict):
                    data.update(params)
                    yield data
                if isinstance(data, list):
                    for item in data:
                        if isinstance(item, dict):
                            item.update(params)
                            yield item
                        else:
                            err = 'Unsupported data format: data=%s, type=%s' \
                                  % (item, type(item))
                            raise Exception(err)
                bbuf.flush()
                hbuf.flush()


HTTP_PAT = re.compile( \
    "(https|http)://[-A-Za-z0-9_+&@#/%?=~_|!:,.;]*[-A-Za-z0-9+&@#/%=~_|]")


def validate_url(url):
    "Validate URL"
    if HTTP_PAT.match(url):
        return True
    return False


def pycurl_options():
    "Default set of options for pycurl"
    opts = {
        'FOLLOWLOCATION': 1,
        'CONNECTTIMEOUT': 270,
        'MAXREDIRS': 5,
        'NOSIGNAL': 1,
        'TIMEOUT': 270,
        'SSL_VERIFYPEER': False,
        'VERBOSE': 0
    }
    return opts


def cern_sso_cookie(url, fname, cert, ckey):
    "Obtain cern SSO cookie and store it in given file name"
    cmd = 'cern-get-sso-cookie -cert %s -key %s -r -u %s -o %s' \
          % (cert, ckey, url, fname)
    proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True, env=os.environ)
    proc.wait()


def getdata(urls, ckey, cert, headers=None, options=None, num_conn=50, cookie=None):
    """
    Get data for given list of urls, using provided number of connections
    and user credentials
    """

    if not options:
        options = pycurl_options()

    portForwarder = PortForward(8443)

    # Make a queue with urls
    queue = [portForwarder(u) for u in urls if validate_url(u)]

    # Check args
    num_urls = len(queue)
    num_conn = min(num_conn, num_urls)

    # Pre-allocate a list of curl objects
    mcurl = pycurl.CurlMulti()
    mcurl.handles = []
    for _ in range(num_conn):
        curl = pycurl.Curl()
        curl.fp = None
        for key, val in viewitems(options):
            curl.setopt(getattr(pycurl, key), val)
        curl.setopt(pycurl.SSLKEY, ckey)
        curl.setopt(pycurl.SSLCERT, cert)
        mcurl.handles.append(curl)
        if headers:
            curl.setopt(pycurl.HTTPHEADER, \
                        ["%s: %s" % (k, v) for k, v in viewitems(headers)])

    # Main loop
    freelist = mcurl.handles[:]
    num_processed = 0
    while num_processed < num_urls:
        # If there is an url to process and a free curl object,
        # add to multi-stack
        while queue and freelist:
            url = queue.pop(0)
            curl = freelist.pop()
            curl.setopt(pycurl.URL, url.encode('ascii', 'ignore'))
            if cookie and url in cookie:
                curl.setopt(pycurl.COOKIEFILE, cookie[url])
                curl.setopt(pycurl.COOKIEJAR, cookie[url])
            bbuf = BytesIO()
            hbuf = BytesIO()
            curl.setopt(pycurl.WRITEFUNCTION, bbuf.write)
            curl.setopt(pycurl.HEADERFUNCTION, hbuf.write)
            mcurl.add_handle(curl)
            # store some info
            curl.hbuf = hbuf
            curl.bbuf = bbuf
            curl.url = url
        # Run the internal curl state machine for the multi stack
        while True:
            ret, _ = mcurl.perform()
            if ret != pycurl.E_CALL_MULTI_PERFORM:
                break
        # Check for curl objects which have terminated, and add them to the
        # freelist
        while True:
            num_q, ok_list, err_list = mcurl.info_read()
            for curl in ok_list:
                hdrs = decodeBytesToUnicode(curl.hbuf.getvalue())
                data = decompress(decodeBytesToUnicode(curl.bbuf.getvalue()), ResponseHeader(hdrs).getHeader())
                url = curl.url
                curl.bbuf.flush()
                curl.bbuf.close()
                curl.hbuf.close()
                curl.hbuf = None
                curl.bbuf = None
                mcurl.remove_handle(curl)
                freelist.append(curl)
                yield {'url': url, 'data': data, 'headers': hdrs}
            for curl, errno, errmsg in err_list:
                hdrs = curl.hbuf.getvalue()
                data = curl.bbuf.getvalue()
                url = curl.url
                curl.bbuf.flush()
                curl.bbuf.close()
                curl.hbuf.close()
                curl.hbuf = None
                curl.bbuf = None
                mcurl.remove_handle(curl)
                freelist.append(curl)
                yield {'url': url, 'data': None, 'headers': hdrs, \
                       'error': errmsg, 'code': errno}
            num_processed = num_processed + len(ok_list) + len(err_list)
            if num_q == 0:
                break
        # Currently no more I/O is pending, could do something in the meantime
        # (display a progress bar, etc.).
        # We just call select() to sleep until some more data is available.
        mcurl.select(1.0)

    cleanup(mcurl)


def cleanup(mcurl):
    "Clean-up MultiCurl handles"
    for curl in mcurl.handles:
        if curl.hbuf is not None:
            curl.hbuf.close()
            curl.hbuf = None
        if curl.bbuf is not None:
            curl.bbuf.close()
            curl.bbuf = None
        curl.close()
    mcurl.close()



#!/usr/bin/env python
"""
_EventBased_t_

Event based splitting test.
"""
from __future__ import print_function, division

from builtins import range
import os
import random
import threading
import unittest

from WMCore.DAOFactory import DAOFactory
from WMCore.DataStructs.Run import Run
from WMCore.Database.CMSCouch import CouchServer
from WMCore.JobSplitting.SplitterFactory import SplitterFactory
from WMCore.Services.UUIDLib import makeUUID
from WMCore.WMBS.File import File
from WMCore.WMBS.Fileset import Fileset
from WMCore.WMBS.Subscription import Subscription
from WMCore.WMBS.Workflow import Workflow
from WMQuality.TestInitCouchApp import TestInitCouchApp


class EventBasedTest(unittest.TestCase):
    """
    _EventBasedTest_

    Test event based job splitting.
    """

    def setUp(self):
        """
        _setUp_

        Create two subscriptions: One that contains a single file and one that
        contains multiple files.
        """
        self.testInit = TestInitCouchApp(__file__)
        self.testInit.setLogging()
        self.testInit.setDatabaseConnection(destroyAllDatabase=True)
        self.couchUrl = os.environ["COUCHURL"]
        self.couchDBName = "acdc_event_based_t"
        self.testInit.setupCouch(self.couchDBName, "GroupUser", "ACDC")
        self.testInit.setSchema(customModules=["WMCore.WMBS"],
                                useDefault=False)
        couchSever = CouchServer(dburl=self.couchUrl)
        self.couchDB = couchSever.connectDatabase(self.couchDBName)
        self.populateWMBS()
        self.performanceParams = {'timePerEvent': 12,
                                  'memoryRequirement': 2300,
                                  'sizePerEvent': 400}
        self.eventsPerJob = 100

        return

    def tearDown(self):
        """
        _tearDown_

        Clear out WMBS.
        """
        self.testInit.clearDatabase()
        self.testInit.tearDownCouch()
        return

    def populateWMBS(self):
        """
        _populateWMBS_

        Create files and subscriptions in WMBS
        """
        myThread = threading.currentThread()
        daofactory = DAOFactory(package="WMCore.WMBS",
                                logger=myThread.logger,
                                dbinterface=myThread.dbi)

        locationAction = daofactory(classname="Locations.New")
        locationAction.execute(siteName='s1', pnn="T1_US_FNAL_Disk")
        locationAction.execute(siteName='s2', pnn="T2_CH_CERN")
        self.validLocations = ["T1_US_FNAL_Disk", "T2_CH_CERN"]

        testWorkflow = Workflow(spec="spec.xml", owner="Steve",
                                name="wf001", task="Test")
        testWorkflow.create()

        self.multipleFileFileset = Fileset(name="TestFileset1")
        self.multipleFileFileset.create()
        parentFile = File('/parent/lfn/', size=1000, events=100,
                          locations=set(["T1_US_FNAL_Disk"]))
        parentFile.create()
        for _ in range(10):
            newFile = File(makeUUID(), size=1000, events=100,
                           locations=set(["T1_US_FNAL_Disk"]))
            newFile.create()
            newFile.addParent(lfn=parentFile['lfn'])
            self.multipleFileFileset.addFile(newFile)
        self.multipleFileFileset.commit()
        self.multipleFileSubscription = Subscription(fileset=self.multipleFileFileset,
                                                     workflow=testWorkflow,
                                                     split_algo="EventBased",
                                                     type="Processing")
        self.multipleFileSubscription.create()

        self.singleFileFileset = Fileset(name="TestFileset2")
        self.singleFileFileset.create()
        newFile = File("/some/file/name", size=1000, events=100,
                       locations=set(["T1_US_FNAL_Disk"]))
        newFile.create()
        self.singleFileFileset.addFile(newFile)
        self.singleFileFileset.commit()
        self.singleFileSubscription = Subscription(fileset=self.singleFileFileset,
                                                   workflow=testWorkflow,
                                                   split_algo="EventBased",
                                                   type="Processing")
        self.singleFileSubscription.create()

        self.multipleSiteFileset = Fileset(name="TestFileset3")
        self.multipleSiteFileset.create()
        for _ in range(5):
            newFile = File(makeUUID(), size=1000, events=100)
            newFile.setLocation("T1_US_FNAL_Disk")
            newFile.create()
            self.multipleSiteFileset.addFile(newFile)
        for _ in range(5):
            newFile = File(makeUUID(), size=1000, events=100)
            newFile.setLocation(["T1_US_FNAL_Disk", "T2_CH_CERN"])
            newFile.create()
            self.multipleSiteFileset.addFile(newFile)
        self.multipleSiteFileset.commit()
        self.multipleSiteSubscription = Subscription(fileset=self.multipleSiteFileset,
                                                     workflow=testWorkflow,
                                                     split_algo="EventBased",
                                                     type="Processing")
        self.multipleSiteSubscription.create()

        return

    def populateACDCCouch(self, numFiles=3, lumisPerJob=4,
                          eventsPerJob=100, numberOfJobs=250, acdcVer=None):
        """
        _populateACDCCouch_

        Create production files in couchDB to test the creation
        of ACDC jobs for the EventBased algorithm.
        acdcVer parameter decides whether the ACDC documents have to be created
        following the new behaviour/changes or not, for more details see:
        See issue: https://github.com/dmwm/WMCore/issues/9126
        """
        # Define some constants
        workflowName = "ACDC_TestEventBased"
        filesetName = "/%s/Production" % workflowName

        lumisPerFile = lumisPerJob * numberOfJobs
        for i in range(numFiles):
            lfn = "MCFakeFile-some-hash-%s" % str(i).zfill(5)
            for j in range(numberOfJobs):
                firstEvent = (i * eventsPerJob * numberOfJobs) + (j * eventsPerJob) + 1
                acdcFile = File(lfn=lfn, size=1024, events=eventsPerJob, locations=self.validLocations,
                                merged=False, first_event=firstEvent)
                if acdcVer == 2:
                    # lumi range is really inclusive (process first and last lumi)
                    run = Run(1, *list(range(1 + j * lumisPerJob + (i * lumisPerFile),
                                        1 + (j + 1) * lumisPerJob + (i * lumisPerFile))))
                else:
                    # bigger lumi range, LastLumi is in most of the cases not processed
                    run = Run(1, *list(range(1 + j * lumisPerJob + (i * lumisPerFile),
                                        2 + (j + 1) * lumisPerJob + (i * lumisPerFile))))

                acdcFile.addRun(run)
                acdcDoc = {"collection_name": workflowName,
                           "collection_type": "ACDC.CollectionTypes.DataCollection",
                           "files": {lfn: acdcFile},
                           "acdc_version": 2 if acdcVer else 1,
                           "fileset_name": filesetName}

                self.couchDB.queue(acdcDoc)

        self.couchDB.commit()
        return

    def populateACDCFakeFile(self, acdcVer=1):
        """
        _populateACDCFakeFile_
        Create an ACDC Collection with a MCFakeFile and 5 jobs:
        * 4 jobs with single lumi section
        * last job with 4 lumi sections (or 5 if it's the old ACDC doc)
        """
        eventsPerJob = 700
        workflowName = "ACDC_TestEventBased"
        filesetName = "/%s/Production" % workflowName
        acdcFiles = []
        # create 4 MCFakeFiles with a single lumi section
        for lumi in [337, 197, 529, 421]:
            lfn = "MCFakeFile-some-hash-00000"
            firstEvent = random.randint(1, 100) * eventsPerJob + 1
            acdcFile = File(lfn=lfn, size=1024, events=eventsPerJob, locations=self.validLocations,
                            merged=False, first_event=firstEvent)
            if acdcVer == 2:
                run = Run(1, *list(range(lumi, lumi + 1)))
            else:
                run = Run(1, *list(range(lumi, lumi + 2)))
            acdcFile.addRun(run)
            acdcFiles.append(acdcFile)
        # create one last entry with more lumis
        acdcFile = File(lfn=lfn, size=1024, events=eventsPerJob, locations=self.validLocations,
                        merged=False, first_event=firstEvent)
        if acdcVer == 2:
            run = Run(1, *list(range(277, 277 + 4)))
        else:
            run = Run(1, *list(range(277, 277 + 5)))
        acdcFile.addRun(run)
        acdcFiles.append(acdcFile)

        for f in acdcFiles:
            acdcDoc = {"collection_name": workflowName,
                       "collection_type": "ACDC.CollectionTypes.DataCollection",
                       "files": {f['lfn']: f},
                       "acdc_version": acdcVer,
                       "fileset_name": filesetName}
            self.couchDB.queue(acdcDoc)

        self.couchDB.commit()
        return

    def generateFakeMCFile(self, numEvents=100, firstEvent=1,
                           lastEvent=100, firstLumi=1, lastLumi=10,
                           index=1, existingSub=None):
        """
        _generateFakeMCFile_

        Generates a fake MC file for testing production EventBased
        creation of jobs, it creates a single file subscription if no
        existing subscription is provided.
        """
        # MC comes with MCFakeFile(s)
        newFile = File("MCFakeFile-some-hash-%s" % str(index).zfill(5), size=1000,
                       events=numEvents,
                       locations=set(["T1_US_FNAL_Disk"]))
        newFile.addRun(Run(1, *list(range(firstLumi, lastLumi + 1))))
        newFile["first_event"] = firstEvent
        newFile["last_event"] = lastEvent
        newFile.create()
        if existingSub is None:
            singleMCFileset = Fileset(name="MCTestFileset-%i" % index)
            singleMCFileset.create()
            singleMCFileset.addFile(newFile)
            singleMCFileset.commit()
            testWorkflow = Workflow(spec="spec.xml", owner="Steve",
                                    name="wf001", task="Test")
            testWorkflow.create()
            singleMCFileSubscription = Subscription(fileset=singleMCFileset,
                                                    workflow=testWorkflow,
                                                    split_algo="EventBased",
                                                    type="Production")
            singleMCFileSubscription.create()
            return singleMCFileSubscription
        else:
            existingSub['fileset'].addFile(newFile)
            existingSub['fileset'].commit()
            return existingSub

    def testExactEvents(self):
        """
        _testExactEvents_

        Test event based job splitting when the number of events per job is
        exactly the same as the number of events in the input file.
        """
        splitter = SplitterFactory()
        jobFactory = splitter(package="WMCore.WMBS",
                              subscription=self.singleFileSubscription)

        jobGroups = jobFactory(events_per_job=self.eventsPerJob,
                               performance=self.performanceParams)

        self.assertEqual(len(jobGroups), 1)

        self.assertEqual(len(jobGroups[0].jobs), 1)

        job = jobGroups[0].jobs.pop()

        self.assertEqual(job.getFiles(type="lfn"), ["/some/file/name"])

        self.assertEqual(job["mask"].getMaxEvents(), self.eventsPerJob)
        self.assertEqual(job["mask"]["FirstEvent"], 0)
        self.assertEqual(job["mask"]["LastEvent"], 99)

        self.assertEqual(job["estimatedJobTime"], 100 * 12)
        self.assertEqual(job["estimatedDiskUsage"], 400 * 100)
        self.assertEqual(job["estimatedMemoryUsage"], 2300)

        return

    def testMoreEvents(self):
        """
        _testMoreEvents_

        Test event based job splitting when the number of events per job is
        greater than the number of events in the input file.
        Since the file has less events than the splitting, the job goes without a mask.
        """
        splitter = SplitterFactory()
        jobFactory = splitter(package="WMCore.WMBS", subscription=self.singleFileSubscription)

        jobGroups = jobFactory(events_per_job=1000,
                               performance=self.performanceParams)

        self.assertEqual(len(jobGroups), 1)

        self.assertEqual(len(jobGroups[0].jobs), 1)

        job = jobGroups[0].jobs.pop()

        self.assertEqual(job.getFiles(type="lfn"), ["/some/file/name"])

        self.assertEqual(job["mask"].getMaxEvents(), 100)
        self.assertEqual(job["mask"]["FirstEvent"], 0)
        self.assertEqual(job["mask"]["LastEvent"], 99)

        self.assertEqual(job["estimatedJobTime"], 100 * 12)
        self.assertEqual(job["estimatedDiskUsage"], 400 * 100)
        self.assertEqual(job["estimatedMemoryUsage"], 2300)

        return

    def test50EventSplit(self):
        """
        _test50EventSplit_

        Test event based job splitting when the number of events per job is
        50, this should result in two jobs.
        """
        splitter = SplitterFactory()

        jobFactory = splitter(package="WMCore.WMBS", subscription=self.singleFileSubscription)

        jobGroups = jobFactory(events_per_job=50,
                               performance=self.performanceParams)

        self.assertEqual(len(jobGroups), 1)

        self.assertEqual(len(jobGroups[0].jobs), 2)

        for job in jobGroups[0].jobs:
            self.assertEqual(job.getFiles(type="lfn"), ["/some/file/name"])
            self.assertEqual(job["mask"].getMaxEvents(), 50)

            if job["mask"]["FirstEvent"] == 0:
                self.assertEqual(job["mask"]["LastEvent"], 49)
            elif job["mask"]["FirstEvent"] == 50:
                self.assertEqual(job["mask"]["LastEvent"], 99)
            else:
                self.fail("Unexpected splitting was performed")
            self.assertEqual(job["estimatedJobTime"], 50 * 12)
            self.assertEqual(job["estimatedDiskUsage"], 400 * 50)
            self.assertEqual(job["estimatedMemoryUsage"], 2300)

        return

    def test99EventSplit(self):
        """
        _test99EventSplit_

        Test event based job splitting when the number of events per job is
        99, this should result in two jobs. Last job shouldn't have a maximum
        number of events, let it run until the end of the file.
        """
        splitter = SplitterFactory()
        jobFactory = splitter(package="WMCore.WMBS", subscription=self.singleFileSubscription)

        jobGroups = jobFactory(events_per_job=99,
                               performance=self.performanceParams)

        self.assertEqual(len(jobGroups), 1)

        self.assertEqual(len(jobGroups[0].jobs), 2)

        for job in jobGroups[0].jobs:
            self.assertEqual(job.getFiles(type="lfn"), ["/some/file/name"])

            if job["mask"].getMaxEvents() == 99:
                self.assertEqual(job["mask"]["FirstEvent"], 0)
                self.assertEqual(job["mask"]["LastEvent"], 98)
                self.assertEqual(job["estimatedJobTime"], 99 * 12)
                self.assertEqual(job["estimatedDiskUsage"], 400 * 99)
                self.assertEqual(job["estimatedMemoryUsage"], 2300)
            elif job["mask"].getMaxEvents() == 1:
                self.assertEqual(job["mask"]["FirstEvent"], 99)
                self.assertEqual(job["mask"]["LastEvent"], 99)
                self.assertEqual(job["estimatedJobTime"], 1 * 12)
                self.assertEqual(job["estimatedDiskUsage"], 400 * 1)
                self.assertEqual(job["estimatedMemoryUsage"], 2300)
            else:
                self.fail("Unexpected splitting was performed")
        return

    def test100EventMultipleFileSplit(self):
        """
        _test100EventMultipleFileSplit_

        Test job splitting into 100 event jobs when the input subscription has
        more than one file available.
        """
        splitter = SplitterFactory()
        jobFactory = splitter(package="WMCore.WMBS", subscription=self.multipleFileSubscription)

        jobGroups = jobFactory(events_per_job=self.eventsPerJob,
                               performance=self.performanceParams)

        self.assertEqual(len(jobGroups), 1)

        self.assertEqual(len(jobGroups[0].jobs), 10)

        for job in jobGroups[0].jobs:
            self.assertEqual(len(job.getFiles(type="lfn")), 1)
            self.assertEqual(job["mask"].getMaxEvents(), self.eventsPerJob)
            self.assertEqual(job["mask"]["FirstEvent"], 0)
            self.assertEqual(job["estimatedJobTime"], 100 * 12)
            self.assertEqual(job["estimatedDiskUsage"], 400 * 100)
            self.assertEqual(job["estimatedMemoryUsage"], 2300)

        return

    def test50EventMultipleFileSplit(self):
        """
        _test50EventMultipleFileSplit_

        Test job splitting into 50 event jobs when the input subscription has
        more than one file available.
        """
        splitter = SplitterFactory()
        jobFactory = splitter(package="WMCore.WMBS", subscription=self.multipleFileSubscription)

        jobGroups = jobFactory(events_per_job=50,
                               performance=self.performanceParams)

        self.assertEqual(len(jobGroups), 1)

        self.assertEqual(len(jobGroups[0].jobs), 20)

        for job in jobGroups[0].jobs:
            self.assertEqual(len(job.getFiles(type="lfn")), 1)
            self.assertEqual(job["mask"].getMaxEvents(), 50)

            if job["mask"]["FirstEvent"] == 0:
                self.assertEqual(job["mask"]["LastEvent"], 49)
            elif job["mask"]["FirstEvent"] == 50:
                self.assertEqual(job["mask"]["LastEvent"], 99)
            else:
                self.fail("Unexpected splitting was performed")
            self.assertEqual(job["estimatedJobTime"], 50 * 12)
            self.assertEqual(job["estimatedDiskUsage"], 400 * 50)
            self.assertEqual(job["estimatedMemoryUsage"], 2300)
        return

    def test150EventMultipleFileSplit(self):
        """
        _test150EventMultipleFileSplit_

        Test job splitting into 150 event jobs when the input subscription has
        more than one file available.  This test verifies that the job splitting
        code will put at most one file in a job. Since every job has less events
        than the maximum. the job goes without a mask.
        """
        splitter = SplitterFactory()
        jobFactory = splitter(package="WMCore.WMBS", subscription=self.multipleFileSubscription)

        jobGroups = jobFactory(events_per_job=150,
                               performance=self.performanceParams)

        self.assertEqual(len(jobGroups), 1)
        self.assertEqual(len(jobGroups[0].jobs), 10)

        for job in jobGroups[0].jobs:
            self.assertEqual(len(job.getFiles(type="lfn")), 1)
            self.assertEqual(job["mask"].getMaxEvents(), 100)
            self.assertEqual(job["mask"]["FirstEvent"], 0)
            self.assertEqual(job["mask"]["LastEvent"], 99)
            self.assertEqual(job["estimatedJobTime"], 100 * 12)
            self.assertEqual(job["estimatedDiskUsage"], 400 * 100)
            self.assertEqual(job["estimatedMemoryUsage"], 2300)

        return

    def test100EventMultipleSite(self):
        """
        _test100EventMultipleSite_

        Test job splitting into 100 event jobs when the input subscription has
        more than one file available, at different site combinations.
        """
        splitter = SplitterFactory()
        jobFactory = splitter(package="WMCore.WMBS",
                              subscription=self.multipleSiteSubscription)

        jobGroups = jobFactory(events_per_job=self.eventsPerJob,
                               performance=self.performanceParams)

        self.assertEqual(len(jobGroups), 2)

        self.assertEqual(len(jobGroups[0].jobs), 5)
        self.assertEqual(len(jobGroups[1].jobs), 5)
        self.assertEqual(jobGroups[0].jobs[0]['possiblePSN'], set(['s1', 's2']))
        self.assertEqual(jobGroups[1].jobs[0]['possiblePSN'], set(['s1']))

        for job in jobGroups[0].jobs:
            self.assertEqual(len(job.getFiles(type="lfn")), 1)

            self.assertEqual(job["mask"].getMaxEvents(), self.eventsPerJob)

            self.assertEqual(job["mask"]["FirstEvent"], 0)
            self.assertEqual(job["estimatedJobTime"], 100 * 12)
            self.assertEqual(job["estimatedDiskUsage"], 400 * 100)
            self.assertEqual(job["estimatedMemoryUsage"], 2300)

        return

    def testMCEventSplitOver32bit(self):
        """
        _testMCEventSplitOver32bit_

        Make sure that no events will go over a 32 bit unsigned integer
        representation, event counter should be reset in that case.
        Also test is not over cautious.
        """
        firstEvent = 3 * (2 ** 30) + 1
        singleMCSubscription = self.generateFakeMCFile(numEvents=2 ** 30,
                                                       firstEvent=firstEvent)
        splitter = SplitterFactory()
        jobFactory = splitter(package="WMCore.WMBS", subscription=singleMCSubscription)

        jobGroups = jobFactory(events_per_job=2 ** 30 - 1,
                               events_per_lumi=2 ** 30 - 1,
                               performance=self.performanceParams)
        self.assertEqual(len(jobGroups), 1,
                         "Error: JobFactory did not return one JobGroup")
        self.assertEqual(len(jobGroups[0].jobs), 2,
                         "Error: JobFactory created %s jobs not two"
                         % len(jobGroups[0].jobs))
        for job in jobGroups[0].jobs:

            if job["mask"].getMaxEvents() == 2 ** 30 - 1:
                self.assertEqual(job["mask"]["FirstLumi"], 1)
                self.assertEqual(job["mask"]["FirstEvent"], firstEvent)
                self.assertTrue(job["mask"]["LastEvent"] <= 2 ** 32)
                self.assertEqual(job["estimatedJobTime"], (2 ** 30 - 1) * 12)
                self.assertEqual(job["estimatedDiskUsage"], 400 * (2 ** 30 - 1))
                self.assertEqual(job["estimatedMemoryUsage"], 2300)
            elif job["mask"].getMaxEvents() == 1:
                self.assertEqual(job["mask"]["FirstLumi"], 2)
                self.assertEqual(job["mask"]["FirstEvent"], 1)
                self.assertEqual(job["estimatedJobTime"], 1 * 12)
                self.assertEqual(job["estimatedDiskUsage"], 400 * 1)
                self.assertEqual(job["estimatedMemoryUsage"], 2300)
            else:
                self.fail("Unexpected splitting was performed")

    def test_addParents(self):
        """
        _addParents_

        Test our ability to add parents to a job
        """

        splitter = SplitterFactory()
        jobFactory = splitter(package="WMCore.WMBS",
                              subscription=self.multipleFileSubscription)

        jobGroups = jobFactory(events_per_job=50, include_parents=True,
                               performance=self.performanceParams)

        self.assertEqual(len(jobGroups), 1)

        self.assertEqual(len(jobGroups[0].jobs), 20)

        for job in jobGroups[0].jobs:
            self.assertEqual(len(job.getFiles(type="lfn")), 1)

            self.assertEqual(job["mask"].getMaxEvents(), 50)
            if job["mask"]["FirstEvent"] == 0:
                self.assertEqual(job["mask"]["LastEvent"], 49)
            elif job["mask"]["FirstEvent"] == 50:
                self.assertEqual(job["mask"]["LastEvent"], 99)
            else:
                self.fail("Unexpected splitting was performed")

            self.assertEqual(job["estimatedJobTime"], 50 * 12)
            self.assertEqual(job["estimatedDiskUsage"], 400 * 50)
            self.assertEqual(job["estimatedMemoryUsage"], 2300)
            for f in job['input_files']:
                self.assertEqual(len(f['parents']), 1)
                self.assertEqual(list(f['parents'])[0]['lfn'], '/parent/lfn/')

        return

    def testACDCProduction_v1(self):
        """
        _testACDCProduction_v1_

        Test the ability of the EventBased algorithm of creating
        jobs from ACDC correctly. Uses ACDC documents not versioned.
        """
        numFiles = 3
        lumisPerJob = 4
        eventsPerJob = 100
        numberOfJobs = 20
        self.populateACDCCouch(numFiles=numFiles, lumisPerJob=lumisPerJob,
                               eventsPerJob=eventsPerJob, numberOfJobs=numberOfJobs)

        mcSubscription = None
        for idx in range(3):
            mcSubscription = self.generateFakeMCFile(numEvents=eventsPerJob * numberOfJobs,
                                                     firstEvent=idx * eventsPerJob * numberOfJobs + 1,
                                                     lastEvent=(idx + 1) * eventsPerJob * numberOfJobs,
                                                     firstLumi=idx * lumisPerJob * numberOfJobs + 1,
                                                     lastLumi=(idx + 1) * lumisPerJob * numberOfJobs,
                                                     index=idx, existingSub=mcSubscription)

        splitter = SplitterFactory()
        jobFactory = splitter(package="WMCore.WMBS",
                              subscription=mcSubscription)

        jobGroups = jobFactory(events_per_job=eventsPerJob, events_per_lumi=int(eventsPerJob / lumisPerJob),
                               collectionName="ACDC_TestEventBased",
                               couchURL=self.couchUrl, couchDB=self.couchDBName,
                               filesetName="/ACDC_TestEventBased/Production",
                               performance=self.performanceParams)

        self.assertEqual(1, len(jobGroups))
        jobGroup = jobGroups[0]
        self.assertEqual(numFiles * numberOfJobs, len(jobGroup.jobs))

        for job in jobGroup.jobs:
            self.assertEqual(1, len(job["input_files"]))
            mask = job["mask"]
            self.assertEqual(mask.getMaxEvents(), eventsPerJob)
            self.assertEqual(mask.getMax("Event"), eventsPerJob)
            self.assertEqual(mask.getMax("Lumi"), lumisPerJob)
            self.assertEqual(mask.getMax("Run"), 1)
            self.assertEqual(mask["LastLumi"] - mask["FirstLumi"], lumisPerJob - 1)
            self.assertEqual(mask["LastEvent"] - mask["FirstEvent"], eventsPerJob - 1)
            self.assertEqual(mask["runAndLumis"], {})
            self.assertEqual(job["estimatedJobTime"], eventsPerJob * 12)
            self.assertEqual(job["estimatedDiskUsage"], eventsPerJob * 400)
            self.assertEqual(job["estimatedMemoryUsage"], 2300)

        return

    def testACDCProduction_v2(self):
        """
        _testACDCProduction_v2_

        Test the ability of the EventBased algorithm of creating
        jobs from ACDC correctly. Uses the new ACDC document version.
        """
        numFiles = 3
        lumisPerJob = 4
        eventsPerJob = 100
        numberOfJobs = 12  # 200
        self.populateACDCCouch(numFiles=numFiles, lumisPerJob=lumisPerJob, eventsPerJob=eventsPerJob,
                               numberOfJobs=numberOfJobs, acdcVer=2)

        mcSubscription = None
        for idx in range(3):
            mcSubscription = self.generateFakeMCFile(numEvents=eventsPerJob * numberOfJobs,
                                                     firstEvent=idx * eventsPerJob * numberOfJobs + 1,
                                                     lastEvent=(idx + 1) * eventsPerJob * numberOfJobs,
                                                     firstLumi=idx * lumisPerJob * numberOfJobs + 1,
                                                     lastLumi=(idx + 1) * lumisPerJob * numberOfJobs,
                                                     index=idx, existingSub=mcSubscription)

        splitter = SplitterFactory()
        jobFactory = splitter(package="WMCore.WMBS",
                              subscription=mcSubscription)

        jobGroups = jobFactory(events_per_job=eventsPerJob, events_per_lumi=int(eventsPerJob / lumisPerJob),
                               collectionName="ACDC_TestEventBased",
                               couchURL=self.couchUrl, couchDB=self.couchDBName,
                               filesetName="/ACDC_TestEventBased/Production",
                               performance=self.performanceParams)

        self.assertEqual(1, len(jobGroups))
        jobGroup = jobGroups[0]
        self.assertEqual(numFiles * numberOfJobs, len(jobGroup.jobs))

        for job in jobGroup.jobs:
            self.assertEqual(1, len(job["input_files"]))
            mask = job["mask"]
            self.assertEqual(mask.getMaxEvents(), eventsPerJob)
            self.assertEqual(mask.getMax("Event"), eventsPerJob)
            self.assertEqual(mask.getMax("Lumi"), lumisPerJob)
            self.assertEqual(mask.getMax("Run"), 1)
            self.assertEqual(mask["LastLumi"] - mask["FirstLumi"], lumisPerJob - 1)
            self.assertEqual(mask["LastEvent"] - mask["FirstEvent"], eventsPerJob - 1)
            self.assertEqual(mask["runAndLumis"], {})
            self.assertEqual(job["estimatedJobTime"], eventsPerJob * 12)
            self.assertEqual(job["estimatedDiskUsage"], eventsPerJob * 400)
            self.assertEqual(job["estimatedMemoryUsage"], 2300)

        return

    def testACDCNonSequential_v1(self):
        """
        _testACDCNonSequential_v1_
        Test the ability of the EventBased algorithm to create the proper jobs
        given job information from the ACDCServer using non-sequential and irregular
        (diff number of lumis per job) lumi distribution (old version of ACDC docs)
        """
        eventsPerJob = 700
        eventsPerLumi = 200
        lumisPerJob = 4

        self.populateACDCFakeFile()

        mcSubscription = self.generateFakeMCFile(3500, 1, 0, 1, 1000, 0)
        splitter = SplitterFactory()
        jobFactory = splitter(package="WMCore.WMBS",
                              subscription=mcSubscription)

        jobGroups = jobFactory(events_per_job=eventsPerJob, events_per_lumi=eventsPerLumi,
                               collectionName="ACDC_TestEventBased",
                               couchURL=self.couchUrl, couchDB=self.couchDBName,
                               filesetName="/ACDC_TestEventBased/Production",
                               performance=self.performanceParams)

        self.assertEqual(1, len(jobGroups))
        jobGroup = jobGroups[0]
        self.assertEqual(5, len(jobGroup.jobs))

        for jobNum, lRange in enumerate([[197, 198], [277, 281], [337, 338], [421, 422], [529, 530]]):
            job = jobGroup.jobs[jobNum]
            self.assertEqual(1, len(job["input_files"]))
            mask = job["mask"]
            self.assertEqual(mask.getMaxEvents(), eventsPerJob)
            self.assertEqual(mask.getMax("Event"), eventsPerJob)
            if lRange[0] == 277:
                self.assertEqual(mask.getMax("Lumi"), lumisPerJob)
            else:
                self.assertEqual(mask.getMax("Lumi"), 1)
            self.assertEqual(mask.getMax("Run"), 1)
            self.assertEqual(mask["FirstLumi"], lRange[0])
            self.assertEqual(mask["LastLumi"], lRange[1] - 1)
            self.assertEqual(mask["LastEvent"] - mask["FirstEvent"], eventsPerJob - 1)
            self.assertEqual(mask["runAndLumis"], {})
            self.assertEqual(job["estimatedJobTime"], eventsPerJob * 12)
            self.assertEqual(job["estimatedDiskUsage"], eventsPerJob * 400)
            self.assertEqual(job["estimatedMemoryUsage"], 2300)

        return

    def testACDCNonSequential_v2(self):
        """
        _testACDCNonSequential_v2_
        Test the ability of the EventBased algorithm to create the proper jobs
        given job information from the ACDCServer using non-sequential and irregular
        (diff number of lumis per job) lumi distribution (new version of ACDC docs)
        """
        eventsPerJob = 700
        eventsPerLumi = 200
        lumisPerJob = 4

        self.populateACDCFakeFile(acdcVer=2)

        mcSubscription = self.generateFakeMCFile(3500, 1, 0, 1, 1000, 0)
        splitter = SplitterFactory()
        jobFactory = splitter(package="WMCore.WMBS",
                              subscription=mcSubscription)

        jobGroups = jobFactory(events_per_job=eventsPerJob, events_per_lumi=eventsPerLumi,
                               collectionName="ACDC_TestEventBased",
                               couchURL=self.couchUrl, couchDB=self.couchDBName,
                               filesetName="/ACDC_TestEventBased/Production",
                               performance=self.performanceParams)

        self.assertEqual(1, len(jobGroups))
        jobGroup = jobGroups[0]
        self.assertEqual(5, len(jobGroup.jobs))

        for jobNum, lRange in enumerate([[197, 198], [277, 281], [337, 338], [421, 422], [529, 530]]):
            job = jobGroup.jobs[jobNum]
            self.assertEqual(1, len(job["input_files"]))
            mask = job["mask"]
            self.assertEqual(mask.getMaxEvents(), eventsPerJob)
            self.assertEqual(mask.getMax("Event"), eventsPerJob)
            if lRange[0] == 277:
                self.assertEqual(mask.getMax("Lumi"), lumisPerJob)
            else:
                self.assertEqual(mask.getMax("Lumi"), 1)
            self.assertEqual(mask.getMax("Run"), 1)
            self.assertEqual(mask["FirstLumi"], lRange[0])
            self.assertEqual(mask["LastLumi"], lRange[1] - 1)
            self.assertEqual(mask["LastEvent"] - mask["FirstEvent"], eventsPerJob - 1)
            self.assertEqual(mask["runAndLumis"], {})
            self.assertEqual(job["estimatedJobTime"], eventsPerJob * 12)
            self.assertEqual(job["estimatedDiskUsage"], eventsPerJob * 400)
            self.assertEqual(job["estimatedMemoryUsage"], 2300)

        return


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
"""
Unittests for PortForward
"""

from __future__ import division, print_function

import unittest

from Utils.PortForward import portForward, PortForward


class RequestHandler(object):
    def __init__(self, config=None, logger=None):
        super(RequestHandler, self).__init__()
        if not config:
            config = {}

    @portForward(8443)
    def request(self, url, params=None, headers=None, verb='GET',
                verbose=0, ckey=None, cert=None, doseq=True,
                encode=False, decode=False, cookie=None, uri=None):
        return url


class PortForwardTests(unittest.TestCase):
    """
    Unittest for PortForward decorator and class
    """

    def __init__(self, *args, **kwargs):
        super(PortForwardTests, self).__init__(*args, **kwargs)
        self.urlResultList = []
        self.urlTestList = ['https://cms-service-dqm.web.cern.ch/cms-service-dqm/CAF/certification/Collisions12/8TeV/Reprocessing/Cert_190456-195530_8TeV_08Jun2012ReReco_Collisions12_JSON.txt',
                            'https://cmsweb.cern.ch/couchdb/reqmgr_workload_cache/_design/ReqMgr/_view/bydate?descending=true&limit=1',
                            'https://cmsweb.cern.ch/couchdb/reqmgr_workload_cache/_design/ReqMgr/_view/bystatusandtime?startkey=%5B%22announced%22%2C+0%5D&endkey=%5B%22announced%22%2C+1616016936%5D&descending=false&stale=update_after&include_docs=false',
                            'https://cmsweb.cern.ch:8443/reqmgr2/js/?f=utils.js&f=ajax_utils.js&f=md5.js&f=task_splitting.js',
                            'https://cmsweb.cern.ch:443/wmstatsserver/data/filtered_requests?mask=RequestStatus&mask=RequestType&mask=RequestPriority&mask=Campaign&mask=RequestNumEvents',
                            u'https://cmsweb.cern.ch/couchdb/reqmgr_workload_cache/_design/ReqMgr/_view/bydate?descending=true&limit=1',
                            u'https://cmsweb.cern.ch/couchdb/reqmgr_workload_cache/_design/ReqMgr/_view/bystatusandtime?startkey=%5B%22announced%22%2C+0%5D&endkey=%5B%22announced%22%2C+1616016936%5D&descending=false&stale=update_after&include_docs=false',
                            u'https://cmsweb.cern.ch/reqmgr2/js/?f=utils.js&f=ajax_utils.js&f=md5.js&f=task_splitting.js',
                            u'https://cmsweb.cern.ch/wmstatsserver/data/filtered_requests?mask=RequestStatus&mask=RequestType&mask=RequestPriority&mask=Campaign&mask=RequestNumEvents',
                            b'https://cmsweb.cern.ch/couchdb/reqmgr_workload_cache/_design/ReqMgr/_view/bydate?descending=true&limit=1',
                            b'https://cmsweb.cern.ch/couchdb/reqmgr_workload_cache/_design/ReqMgr/_view/bystatusandtime?startkey=%5B%22announced%22%2C+0%5D&endkey=%5B%22announced%22%2C+1616016936%5D&descending=false&stale=update_after&include_docs=false',
                            b'https://cmsweb.cern.ch/reqmgr2/js/?f=utils.js&f=ajax_utils.js&f=md5.js&f=task_splitting.js',
                            b'https://cmsweb.cern.ch/wmstatsserver/data/filtered_requests?mask=RequestStatus&mask=RequestType&mask=RequestPriority&mask=Campaign&mask=RequestNumEvents']

        self.urlExpectedtList = ['https://cms-service-dqm.web.cern.ch/cms-service-dqm/CAF/certification/Collisions12/8TeV/Reprocessing/Cert_190456-195530_8TeV_08Jun2012ReReco_Collisions12_JSON.txt',
                                 'https://cmsweb.cern.ch:8443/couchdb/reqmgr_workload_cache/_design/ReqMgr/_view/bydate?descending=true&limit=1',
                                 'https://cmsweb.cern.ch:8443/couchdb/reqmgr_workload_cache/_design/ReqMgr/_view/bystatusandtime?startkey=%5B%22announced%22%2C+0%5D&endkey=%5B%22announced%22%2C+1616016936%5D&descending=false&stale=update_after&include_docs=false',
                                 'https://cmsweb.cern.ch:8443/reqmgr2/js/?f=utils.js&f=ajax_utils.js&f=md5.js&f=task_splitting.js',
                                 'https://cmsweb.cern.ch:443/wmstatsserver/data/filtered_requests?mask=RequestStatus&mask=RequestType&mask=RequestPriority&mask=Campaign&mask=RequestNumEvents',
                                 u'https://cmsweb.cern.ch:8443/couchdb/reqmgr_workload_cache/_design/ReqMgr/_view/bydate?descending=true&limit=1',
                                 u'https://cmsweb.cern.ch:8443/couchdb/reqmgr_workload_cache/_design/ReqMgr/_view/bystatusandtime?startkey=%5B%22announced%22%2C+0%5D&endkey=%5B%22announced%22%2C+1616016936%5D&descending=false&stale=update_after&include_docs=false',
                                 u'https://cmsweb.cern.ch:8443/reqmgr2/js/?f=utils.js&f=ajax_utils.js&f=md5.js&f=task_splitting.js',
                                 u'https://cmsweb.cern.ch:8443/wmstatsserver/data/filtered_requests?mask=RequestStatus&mask=RequestType&mask=RequestPriority&mask=Campaign&mask=RequestNumEvents',
                                 b'https://cmsweb.cern.ch:8443/couchdb/reqmgr_workload_cache/_design/ReqMgr/_view/bydate?descending=true&limit=1',
                                 b'https://cmsweb.cern.ch:8443/couchdb/reqmgr_workload_cache/_design/ReqMgr/_view/bystatusandtime?startkey=%5B%22announced%22%2C+0%5D&endkey=%5B%22announced%22%2C+1616016936%5D&descending=false&stale=update_after&include_docs=false',
                                 b'https://cmsweb.cern.ch:8443/reqmgr2/js/?f=utils.js&f=ajax_utils.js&f=md5.js&f=task_splitting.js',
                                 b'https://cmsweb.cern.ch:8443/wmstatsserver/data/filtered_requests?mask=RequestStatus&mask=RequestType&mask=RequestPriority&mask=Campaign&mask=RequestNumEvents']

    def testDecorator(self):
        requesHandler = RequestHandler()
        self.urlResultList = []
        for url in self.urlTestList:
            self.urlResultList.append(requesHandler.request(url))
        self.assertListEqual(self.urlResultList, self.urlExpectedtList)

    def testCallClass(self):
        portForwarder = PortForward(8443)
        self.urlResultList = []
        for url in self.urlTestList:
            self.urlResultList.append(portForwarder(url))
        self.assertListEqual(self.urlResultList, self.urlExpectedtList)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# -*- coding: ISO-8859-1 -*-
"""
Created on Aug 6, 2009

@author: meloam
"""
from __future__ import print_function
from future import standard_library
standard_library.install_aliases()

import json
import os
import shutil
import tempfile
import time
import unittest
from http.client import HTTPException

import nose

import WMCore.Services.Requests as Requests
from WMCore.DataStructs.Job import Job as DataStructsJob
from WMCore.DataStructs.Mask import Mask
from WMCore.DataStructs.Run import Run
from WMCore.Services.Requests import JSONRequests
from WMCore.WMInit import getWMBASE
from WMQuality.TestInit import TestInit
from WMQuality.WebTools.RESTBaseUnitTest import RESTBaseUnitTest
from WMQuality.WebTools.RESTServerSetup import DefaultConfig


class testRequestExceptions(unittest.TestCase):
    def setUp(self):
        self.tmp = tempfile.mkdtemp()
        self.request_dict = {'req_cache_path': self.tmp}

    def tearDown(self):
        shutil.rmtree(self.tmp, ignore_errors=True)

    def test404Error(self):
        endp = "http://cmsweb.cern.ch"
        url = "/thispagedoesntexist/"
        req = Requests.Requests(endp, self.request_dict)
        for v in ['GET', 'POST']:
            self.assertRaises(HTTPException, req.makeRequest, url, verb=v)
        try:
            req.makeRequest(url, verb='GET')
        except HTTPException as e:
            # print e
            self.assertEqual(e.status, 404)

    def test404Error_with_pycurl(self):
        endp = "http://cmsweb.cern.ch"
        url = "/thispagedoesntexist/"
        idict = dict(self.request_dict)
        idict.update({'pycurl': 1})
        req = Requests.Requests(endp, idict)
        for v in ['GET', 'POST']:
            self.assertRaises(HTTPException, req.makeRequest, url, verb=v)
        try:
            req.makeRequest(url, verb='GET')
        except HTTPException as e:
            # print e
            self.assertEqual(e.status, 404)


# comment out so we don't ddos someone else's server
#    def test408Error(self):
#        endp = "http://bitworking.org/projects/httplib2/test"
#        url = "/timeout/timeout.cgi"
#        self.request_dict['timeout'] = 1
#        req = Requests.Requests(endp, self.request_dict) # takes 3 secs to respond
#        self.assertRaises(HTTPException, req.makeRequest, url, verb='GET',
#                          incoming_headers = {'cache-control': 'no-cache'})



class testRepeatCalls(RESTBaseUnitTest):
    def initialize(self):
        self.config = DefaultConfig()
        self.config.UnitTests.templates = getWMBASE() + '/src/templates/WMCore/WebTools'
        self.config.Webtools.section_('server')
        self.config.Webtools.server.socket_timeout = 1
        self.urlbase = self.config.getServerUrl()
        self.cache_path = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.cache_path, ignore_errors=True)
        self.rt.stop()

    def test10Calls(self):
        fail_count = 0
        req = Requests.Requests(self.urlbase, {'req_cache_path': self.cache_path})

        for i in range(0, 5):
            time.sleep(i)
            print('test %s starting at %s' % (i, time.time()))
            try:
                result = req.get('/', incoming_headers={'Cache-Control': 'no-cache'})
                self.assertEqual(False, result[3])
                self.assertEqual(200, result[1])
            except HTTPException as he:
                print('test %s raised a %s error' % (i, he.status))
                fail_count += 1
            except Exception as e:
                print('test %s raised an unexpected exception of type %s' % (i, type(e)))
                print(e)
                fail_count += 1
        if fail_count > 0:
            raise Exception('Test did not pass!')

    def test10Calls_with_pycurl(self):
        fail_count = 0
        idict = {'req_cache_path': self.cache_path, 'pycurl': 1}
        req = Requests.Requests(self.urlbase, idict)

        for i in range(0, 5):
            time.sleep(i)
            print('test %s starting at %s' % (i, time.time()))
            try:
                result = req.get('/', incoming_headers={'Cache-Control': 'no-cache'}, decode=True)
                self.assertEqual(False, result[3])
                self.assertEqual(200, result[1])
            except HTTPException as he:
                print('test %s raised a %s error' % (i, he.status))
                fail_count += 1
            except Exception as e:
                print('test %s raised an unexpected exception of type %s' % (i, type(e)))
                print(e)
                fail_count += 1
        if fail_count > 0:
            raise Exception('Test did not pass!')

    def testRecoveryFromConnRefused(self):
        """Connections succeed after server down"""
        import socket
        self.rt.stop()
        req = Requests.Requests(self.urlbase, {'req_cache_path': self.cache_path, 'pycurl': False})
        headers = {'Cache-Control': 'no-cache'}
        self.assertRaises(socket.error, req.get, '/', incoming_headers=headers)

        # now restart server and hope we can connect
        self.rt.start(blocking=False)
        result = req.get('/', incoming_headers=headers)
        self.assertEqual(result[3], False)
        self.assertEqual(result[1], 200)

    def testRecoveryFromConnRefused_with_pycurl(self):
        """Connections succeed after server down"""
        import pycurl
        self.rt.stop()
        idict = {'req_cache_path': self.cache_path, 'pycurl': 1}
        req = Requests.Requests(self.urlbase, idict)
        headers = {'Cache-Control': 'no-cache'}
        self.assertRaises(pycurl.error, req.get, '/', incoming_headers=headers, decode=True)

        # now restart server and hope we can connect
        self.rt.start(blocking=False)
        result = req.get('/', incoming_headers=headers, decode=True)
        self.assertEqual(result[3], False)
        self.assertEqual(result[1], 200)


class testJSONRequests(unittest.TestCase):
    def setUp(self):
        self.testInit = TestInit(__file__)
        self.testInit.setLogging()
        tmp = self.testInit.generateWorkDir()
        self.request = Requests.JSONRequests(idict={'req_cache_path': tmp})

    def roundTrip(self, data):
        encoded = self.request.encode(data)
        # print encoded
        # print encoded.__class__.__name__
        decoded = self.request.decode(encoded)
        # print decoded.__class__.__name__
        self.assertEqual(data, decoded)

    def roundTripLax(self, data):
        encoded = self.request.encode(data)
        decoded = self.request.decode(encoded)
        datakeys = list(data.keys())

        for k in decoded.keys():
            assert k in datakeys
            datakeys.pop(datakeys.index(k))
            # print 'the following keys were dropped\n\t',datakeys

    def testSet1(self):
        self.roundTrip(set([]))

    def testSet2(self):
        self.roundTrip(set([1, 2, 3, 4, Run(1)]))

    def testSet3(self):
        self.roundTrip(set(['a', 'b', 'c', 'd']))

    def testSet4(self):
        self.roundTrip(set([1, 2, 3, 4, 'a', 'b']))

    def testRun1(self):
        self.roundTrip(Run(1))

    def testRun2(self):
        self.roundTrip(Run(1, 1))

    def testRun3(self):
        self.roundTrip(Run(1, 2, 3))

    def testMask1(self):
        self.roundTrip(Mask())

    def testMask2(self):
        mymask = Mask()
        mymask['FirstEvent'] = 9999
        mymask['LastEvent'] = 999
        self.roundTrip(mymask)

    def testMask3(self):
        mymask = Mask()
        mymask['FirstEvent'] = 9999
        mymask['LastEvent'] = 999
        myjob = DataStructsJob()
        myjob["mask"] = mymask
        self.roundTrip(myjob)

    def testMask4(self):
        self.roundTrip({'LastRun': None, 'FirstRun': None, 'LastEvent': None,
                        'FirstEvent': None, 'LastLumi': None, 'FirstLumi': None})

    def testMask5(self):
        mymask = Mask()
        mymask['FirstEvent'] = 9999
        mymask['LastEvent'] = 999
        myjob = DataStructsJob()
        myjob["mask"] = mymask
        self.roundTripLax(myjob)

    def testMask6(self):
        mymask = Mask()
        myjob = DataStructsJob()
        myjob["mask"] = mymask
        self.roundTripLax(myjob)

    def testSpecialCharacterPasswords(self):
        url = 'http://username:p@ssw:rd@localhost:6666'
        req = JSONRequests(url)
        # the url is sanitized before persisted in the object
        self.assertEqual(req['host'], "http://localhost:6666")
        self.assertEqual(req.additionalHeaders['Authorization'], 'Basic dXNlcm5hbWU6cEBzc3c6cmQ=')


class TestRequests(unittest.TestCase):
    def testGetKeyCert(self):
        """test existance of key/cert"""
        proxy = os.environ.get('X509_USER_PROXY')
        if not proxy:
            raise nose.SkipTest('Only run if an X509 proxy is present')
        os.environ.pop('X509_HOST_CERT', None)
        os.environ.pop('X509_HOST_KEY', None)
        os.environ.pop('X509_USER_CERT', None)
        os.environ.pop('X509_USER_KEY', None)
        req = Requests.Requests('https://cmsweb.cern.ch')
        key, cert = req.getKeyCert()
        self.assertNotEqual(None, key)
        self.assertNotEqual(None, cert)

    def testSecureWithProxy(self):
        """https with proxy"""
        proxy = os.environ.get('X509_USER_PROXY')
        if not proxy:
            raise nose.SkipTest('Only run if an X509 proxy is present')
        os.environ.pop('X509_HOST_CERT', None)
        os.environ.pop('X509_HOST_KEY', None)
        os.environ.pop('X509_USER_CERT', None)
        os.environ.pop('X509_USER_KEY', None)
        req = Requests.Requests('https://cmsweb.cern.ch')
        out = req.makeRequest('/auth/trouble')
        self.assertEqual(out[1], 200)
        self.assertNotEqual(out[0].find('passed basic validation'), -1)
        self.assertNotEqual(out[0].find('certificate is a proxy'), -1)

    def testSecureWithProxy_with_pycurl(self):
        """https with proxy with pycurl"""
        proxy = os.environ.get('X509_USER_PROXY')
        if not proxy:
            raise nose.SkipTest('Only run if an X509 proxy is present')
        os.environ.pop('X509_HOST_CERT', None)
        os.environ.pop('X509_HOST_KEY', None)
        os.environ.pop('X509_USER_CERT', None)
        os.environ.pop('X509_USER_KEY', None)
        req = Requests.Requests('https://cmsweb.cern.ch', {'pycurl': 1})
        out = req.makeRequest('/phedex/datasvc/json/prod/groups', decoder=True)
        self.assertEqual(out[1], 200)
        if not isinstance(json.loads(out[0]), dict):
            msg = 'wrong data type'
            raise Exception(msg)
        out = req.makeRequest('/auth/trouble', decoder=True)
        self.assertEqual(out[1], 200)
        self.assertNotEqual(out[0].find('passed basic validation'), -1)
        self.assertNotEqual(out[0].find('certificate is a proxy'), -1)

    def testSecureNoAuth(self):
        """https with no client authentication"""
        req = Requests.Requests('https://cmsweb.cern.ch')
        out = req.makeRequest('')
        self.assertEqual(out[1], 200)
        # we should get an html page in response
        self.assertNotEqual(out[0].find('html'), -1)

    def testSecureNoAuth_with_pycurl(self):
        """https with no client authentication"""
        req = Requests.Requests('https://cmsweb.cern.ch', {'pycurl': 1})
        out = req.makeRequest('', decoder=True)
        self.assertEqual(out[1], 200)
        # we should get an html page in response
        self.assertNotEqual(out[0].find('html'), -1)

    def testSecureOddPort(self):
        """https with odd port"""
        proxy = os.environ.get('X509_USER_PROXY')
        if not proxy:
            raise nose.SkipTest('Only run if an X509 proxy is present')
        os.environ.pop('X509_HOST_CERT', None)
        os.environ.pop('X509_HOST_KEY', None)
        os.environ.pop('X509_USER_CERT', None)
        os.environ.pop('X509_USER_KEY', None)
        req = Requests.Requests('https://cmsweb.cern.ch:443')
        out = req.makeRequest('/auth/trouble')
        self.assertEqual(out[1], 200)
        self.assertNotEqual(out[0].find('passed basic validation'), -1)

    def testSecureOddPort_with_pycurl(self):
        """https with odd port"""
        proxy = os.environ.get('X509_USER_PROXY')
        if not proxy:
            raise nose.SkipTest('Only run if an X509 proxy is present')
        os.environ.pop('X509_HOST_CERT', None)
        os.environ.pop('X509_HOST_KEY', None)
        os.environ.pop('X509_USER_CERT', None)
        os.environ.pop('X509_USER_KEY', None)
        req = Requests.Requests('https://cmsweb.cern.ch:443', {'pycurl': 1})
        out = req.makeRequest('/auth/trouble', decoder=True)
        self.assertEqual(out[1], 200)
        self.assertNotEqual(out[0].find('passed basic validation'), -1)

    def testNoCache(self):
        """Cache disabled"""
        req = Requests.Requests('https://cmssdt.cern.ch/SDT/', {'cachepath': None})
        out = req.makeRequest('/', decoder=True)
        self.assertEqual(out[3], False)
        self.assertTrue('html' in out[0])

if __name__ == "__main__":
    unittest.main()



from builtins import object
import cherrypy
from WMCore.Database.CMSCouch import Database, CouchError
from WMCore.ReqMgr.DataStructs.ReqMgrConfigDataCache import ReqMgrConfigDataCache

def couch_request_error(func):
    def wrapperFunc(*args, **kwargs):
        try:
            func(*args, **kwargs)
        except CouchError as ex:
            #msg = ("ERROR: Query of CouchDB database %s,  couchapp %s view %s failed."
            #        % (db_name, app_name, view_name))
            #cherrypy.log("%s Reason: %s" % (msg, ex))
            msg = "test"
            raise cherrypy.HTTPError(400, "%s Reason: %s" % (msg, ex))
        return wrapperFunc

def couch_err_handler(decorator):
    def decorate(cls):
        for attr in cls.__dict__:
            if callable(getattr(cls, attr)):
                setattr(cls, attr, decorator(getattr(cls, attr)))
        return cls
    return decorate

@couch_err_handler(couch_request_error)
class RESTBackendCouchDB(Database):
    pass


class ReqMgrCouch(object):

    def __init__(self, config):
        self.config = config
        cherrypy.log("Creating CouchDB connection instances to "
                     "CouchDB host: '%s' ..." % config.couch_host)
        # each CouchDB database will be available under it name in this instance
        dbs = (config.couch_reqmgr_db,
               config.couch_reqmgr_aux_db)
               # will be necessary later:
               #config.couch_config_cache_db,
               #config.couch_workload_summary_db,
               #config.couch_wmstats_db,
        for db in dbs:
            setattr(self, db, self._create_conn(db))

        aux_db = self.get_db(config.couch_reqmgr_aux_db)
        ReqMgrConfigDataCache.set_aux_db(aux_db)

    def _create_conn(self, db_name):
        cherrypy.log("Creating CouchDB connection to '%s' ... " % db_name)
        return RESTBackendCouchDB(dbname=db_name, url=self.config.couch_host)


    def get_db(self, db_name):
        """
        Return corresponding instance of CMSCouch.Database already
        created in the constructor.

        """
        try:
            db = getattr(self, db_name)
            return db
        except AttributeError:
            import traceback
            import sys
            trace = traceback.format_exception(*sys.exc_info())
            trace_str = ''.join(trace)
            msg = "Wrong database name '%s':\n%s" % (db_name, trace_str)
            cherrypy.log(msg)
            raise cherrypy.HTTPError(500, "Internal application error.")



